"""BaSyx Server interface for REST API communication."""

import json
import logging
from enum import Enum
from pathlib import Path
from typing import Any

import puremagic
from basyx.aas import model

from aas_http_client.classes.client.aas_client import AasHttpClient, _create_client
from aas_http_client.classes.wrapper.attachment import Attachment
from aas_http_client.classes.wrapper.pagination import (
    ReferencePaginatedData,
    ShellPaginatedData,
    SubmodelElementPaginatedData,
    SubmodelPaginatedData,
    create_reference_paging_data,
    create_shell_paging_data,
    create_submodel_element_paging_data,
    create_submodel_paging_data,
)
from aas_http_client.utilities.sdk_tools import convert_to_dict as _to_dict
from aas_http_client.utilities.sdk_tools import convert_to_object as _to_object

_logger = logging.getLogger(__name__)


class IdEncoding(Enum):
    """Determines the ID encoding mode for API requests."""

    default = 0
    encoded = 1
    decoded = 2

    def __str__(self) -> str:
        """String representation of the IdMode enum."""
        if self == IdEncoding.encoded:
            return "encoded"
        if self == IdEncoding.decoded:
            return "decoded"

        return ""


class Level(Enum):
    """Determines the structural depth of the respective resource content."""

    default = 0
    core = 1
    deep = 2

    def __str__(self) -> str:
        """String representation of the Level enum."""
        if self == Level.core:
            return "core"
        if self == Level.deep:
            return "deep"

        return ""


class Extent(Enum):
    """Determines to which extent the resource is being serialized."""

    default = 0
    with_blob_value = 1
    without_blob_value = 2

    def __str__(self) -> str:
        """String representation of the Extent enum."""
        if self == Extent.with_blob_value:
            return "withBlobValue"
        if self == Extent.without_blob_value:
            return "withoutBlobValue"

        return ""


class AssetKind(Enum):
    """Determines to which asset kind the resource is being serialized."""

    default = 0
    instance = 1
    not_applicable = 2
    type = 3

    def __str__(self) -> str:
        """String representation of the Extent enum."""
        if self == AssetKind.instance:
            return "Instance"
        if self == AssetKind.not_applicable:
            return "NotApplicable"
        if self == AssetKind.type:
            return "Type"

        return ""


# region SdkWrapper


class SdkWrapper:
    """Represents a wrapper for the BaSyx Python SDK to communicate with a REST API."""

    _client: AasHttpClient
    base_url: str = ""

    def __init__(self, config_string: str, basic_auth_password: str = "", o_auth_client_secret: str = "", bearer_auth_token: str = ""):
        """Initializes the wrapper with the given configuration.

        :param config_string: Configuration string for the BaSyx server connection.
        :param basic_auth_password: Password for the BaSyx server interface client, defaults to "".
        :param o_auth_client_secret: Client secret for OAuth authentication, defaults to "".
        :param bearer_auth_token: Bearer token for authentication, defaults to "".
        """
        client = _create_client(config_string, basic_auth_password, o_auth_client_secret, bearer_auth_token)

        if not client:
            raise ValueError("Failed to create AAS HTTP client with the provided configuration.")

        self._client = client
        self.base_url = client.base_url

    def set_encoded_ids(self, encoded_ids: IdEncoding):
        """Sets whether to use encoded IDs for API requests.

        :param encoded_ids: If enabled, all IDs used in API requests have to be base64-encoded
        """
        if encoded_ids == IdEncoding.encoded:
            self._client.encoded_ids = True
        else:
            self._client.encoded_ids = False

    def get_encoded_ids(self) -> IdEncoding:
        """Gets whether encoded IDs are used for API requests.

        :return: True if encoded IDs are used, False otherwise
        """
        if self._client.encoded_ids:
            return IdEncoding.encoded

        return IdEncoding.decoded

    def get_client(self) -> AasHttpClient:
        """Returns the underlying AAS HTTP client.

        :return: The AAS HTTP client instance.
        """
        return self._client

    # endregion

    # region shells

    # GET /shells/{aasIdentifier}
    def get_asset_administration_shell_by_id(self, aas_identifier: str) -> model.AssetAdministrationShell | None:
        """Returns a specific Asset Administration Shell.

        :param aas_identifier: The Asset Administration Shells unique id (decoded)
        :return: Asset Administration Shells or None if an error occurred
        """
        if not self._client.shells:
            _logger.error("Shell API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return None

        content = self._client.shells.get_asset_administration_shell_by_id(aas_identifier)

        if not content:
            _logger.warning(f"No shell found with ID '{aas_identifier}' on server.")
            return None

        return _to_object(content)

    # PUT /shells/{aasIdentifier}
    def put_asset_administration_shell_by_id(self, aas_identifier: str, aas: model.AssetAdministrationShell) -> bool:
        """Creates or replaces an existing Asset Administration Shell.

        :param aas_identifier: The Asset Administration Shells unique id (decoded)
        :param aas: Asset Administration Shell to put
        :return: True if the update was successful, False otherwise
        """
        if not self._client.shells:
            _logger.error("Shell API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return False

        aas_data = _to_dict(aas)

        if aas_data is None:
            _logger.error(f"Failed to serialize Asset Administration Shell with ID '{aas_identifier}' to dictionary.")
            return False

        return self._client.shells.put_asset_administration_shell_by_id(aas_identifier, aas_data)

    # DELETE /shells/{aasIdentifier}
    def delete_asset_administration_shell_by_id(self, aas_identifier: str) -> bool:
        """Deletes an Asset Administration Shell.

        :param aas_identifier: The Asset Administration Shells unique id (decoded)
        :return: True if the deletion was successful, False otherwise
        """
        if not self._client.shells:
            _logger.error("Shell API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return False

        return self._client.shells.delete_asset_administration_shell_by_id(aas_identifier)

    # GET /shells/{aasIdentifier}/asset-information/thumbnail
    def get_thumbnail_aas_repository(self, aas_identifier: str) -> Attachment | None:
        """Downloads the thumbnail of a specific Asset Administration Shell.

        :param aas_identifier: The Asset Administration Shells unique id (decoded)
        :return: Attachment object with thumbnail content as bytes (octet-stream) or None if an error occurred
        """
        if not self._client.shells:
            _logger.error("Shell API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return None

        byte_content = self._client.shells.get_thumbnail_aas_repository(aas_identifier)

        if not byte_content:
            _logger.warning(f"No thumbnail found for AAS with ID '{aas_identifier}' on server.")
            return None

        return Attachment(
            content=byte_content,
            content_type=puremagic.from_string(byte_content, mime=True),
            filename="thumbnail",
        )

    # PUT /shells/{aasIdentifier}/asset-information/thumbnail
    def put_thumbnail_aas_repository(self, aas_identifier: str, file_name: str, file: Path) -> bool:
        """Creates or updates the thumbnail of the Asset Administration Shell.

        :param aas_identifier: The Asset Administration Shells unique id
        :param file_name: The name of the thumbnail file
        :param file: Path to the thumbnail file to upload as attachment
        :return: True if the update was successful, False otherwise
        """
        if not self._client.shells:
            _logger.error("Shell API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return False

        return self._client.shells.put_thumbnail_aas_repository(aas_identifier, file_name, file)

    # DELETE /shells/{aasIdentifier}/asset-information/thumbnail
    def delete_thumbnail_aas_repository(self, aas_identifier: str) -> bool:
        """Deletes the thumbnail of a specific Asset Administration Shell.

        :param aas_identifier: The Asset Administration Shells unique id (decoded)
        :return: True if the deletion was successful, False otherwise
        """
        if not self._client.shells:
            _logger.error("Shell API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return False

        return self._client.shells.delete_thumbnail_aas_repository(aas_identifier)

    # GET /shells
    def get_all_asset_administration_shells(
        self, asset_ids: list[dict] | None = None, id_short: str = "", limit: int = 100, cursor: str = ""
    ) -> ShellPaginatedData | None:
        """Returns all Asset Administration Shells.

        :param assetIds: A list of specific Asset identifiers (format: {"identifier": "string",  "encodedIdentifier": "string"})
        :param idShort: The Asset Administration Shell's IdShort
        :param limit: The maximum number of elements in the response array
        :param cursor: A server-generated identifier retrieved from pagingMetadata that specifies from which position the result listing should continue
        :return: List of paginated Asset Administration Shells or None if an error occurred
        """
        if not self._client.shells:
            _logger.error("Shell API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return None

        content = self._client.shells.get_all_asset_administration_shells(asset_ids, id_short, limit, cursor)

        if not content:
            return None

        return create_shell_paging_data(content)

    # POST /shells
    def post_asset_administration_shell(self, aas: model.AssetAdministrationShell) -> model.AssetAdministrationShell | None:
        """Creates a new Asset Administration Shell.

        :param aas: Asset Administration Shell to post
        :return: Asset Administration Shell or None if an error occurred
        """
        if not self._client.shells:
            _logger.error("Shell API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return None

        aas_data = _to_dict(aas)
        if aas_data is None:
            return None

        content = self._client.shells.post_asset_administration_shell(aas_data)
        if not content:
            return None

        return _to_object(content)

    # GET /shells/{aasIdentifier}/submodel-refs
    def get_all_submodel_references_aas_repository(self, aas_identifier: str, limit: int = 100, cursor: str = "") -> ReferencePaginatedData | None:
        """Returns all submodel references.

        :param aas_identifier: The Asset Administration Shells unique id
        :param limit: The maximum number of elements in the response array
        :param cursor: A server-generated identifier retrieved from pagingMetadata that specifies from which position the result listing should continue
        :return: List of paginated Submodel References or None if an error occurred
        """
        if not self._client.shells:
            _logger.error("Shell API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return None

        references_result = self._client.shells.get_all_submodel_references_aas_repository(aas_identifier, limit, cursor)

        if not references_result:
            return None

        return create_reference_paging_data(references_result)

    # POST /shells/{aasIdentifier}/submodel-refs
    def post_submodel_reference_aas_repository(self, aas_identifier: str, submodel_reference: model.ModelReference) -> model.ModelReference | None:
        """Creates a submodel reference at the Asset Administration Shell.

        :param aas_identifier: The Asset Administration Shells unique id
        :param submodel_reference: Reference to the Submodel
        :return: Reference Submodel object or None if an error occurred
        """
        if not self._client.shells:
            _logger.error("Shell API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return None

        ref_data = _to_dict(submodel_reference)
        if ref_data is None:
            return None

        content = self._client.shells.post_submodel_reference_aas_repository(aas_identifier, ref_data)
        if not content:
            return None

        return _to_object(content)

    # DELETE /shells/{aasIdentifier}/submodel-refs/{submodelIdentifier}
    def delete_submodel_reference_by_id_aas_repository(self, aas_identifier: str, submodel_identifier: str) -> bool:
        """Deletes the submodel reference from the Asset Administration Shell. Does not delete the submodel itself.

        :param aas_identifier: The Asset Administration Shells unique id
        :param submodel_identifier: The Submodels unique id
        :return: True if the deletion was successful, False otherwise
        """
        if not self._client.shells:
            _logger.error("Shell API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return False

        return self._client.shells.delete_submodel_reference_by_id_aas_repository(aas_identifier, submodel_identifier)

    # not supported by Java Server

    # PUT /shells/{aasIdentifier}/submodels/{submodelIdentifier}
    def put_submodel_by_id_aas_repository(self, aas_identifier: str, submodel_identifier: str, submodel: model.Submodel) -> bool:
        """Updates the Submodel.

        :param aas_identifier: The Asset Administration Shells unique id (decoded)
        :param submodel_identifier: ID of the submodel to put
        :param submodel: Submodel to put
        :return: True if the update was successful, False otherwise
        """
        if not self._client.shells:
            _logger.error("Shell API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return False

        sm_data = _to_dict(submodel)

        if sm_data is None:
            return False

        return self._client.shells.put_submodel_by_id_aas_repository(aas_identifier, submodel_identifier, sm_data)

    # GET /shells/{aasIdentifier}/$reference
    def get_asset_administration_shell_by_id_reference_aas_repository(self, aas_identifier: str) -> model.Reference | None:
        """Returns a specific Asset Administration Shell as a Reference.

        :param aas_identifier: ID of the AAS reference to retrieve
        :return: Asset Administration Shells reference object or None if an error occurred
        """
        # workaround because serialization not working
        aas = self.get_asset_administration_shell_by_id(aas_identifier)

        if not aas:
            return None

        return model.ModelReference.from_referable(aas)

    # GET /shells/{aasIdentifier}/submodels/{submodelIdentifier}
    def get_submodel_by_id_aas_repository(self, aas_identifier: str, submodel_identifier: str) -> model.Submodel | None:
        """Returns the Submodel.

        :param aas_identifier: ID of the AAS to retrieve the submodel from
        :param submodel_identifier: ID of the submodel to retrieve
        :return: Submodel or None if an error occurred
        """
        if not self._client.shells:
            _logger.error("Shell API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return None

        content = self._client.shells.get_submodel_by_id_aas_repository(aas_identifier, submodel_identifier)

        if not content:
            return None

        return _to_object(content)

    # endregion

    # region submodels

    # GET /submodels/{submodelIdentifier}
    def get_submodel_by_id(self, submodel_identifier: str, level: Level = Level.default, extent: Extent = Extent.default) -> model.Submodel | None:
        """Returns a specific Submodel.

        :param submodel_identifier: Encoded ID of the Submodel to retrieve
        :param level: Determines the structural depth of the respective resource content. Available values : deep, core
        :param extent: Determines to which extent the resource is being serialized. Available values : withBlobValue, withoutBlobValue
        :return: Submodel data or None if an error occurred
        """
        if not self._client.submodels:
            _logger.error("Submodel API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return None

        content = self._client.submodels.get_submodel_by_id(submodel_identifier, str(level), str(extent))

        if not content:
            return None

        return _to_object(content)

    # PUT /submodels/{submodelIdentifier}
    def put_submodels_by_id(self, submodel_identifier: str, submodel: model.Submodel) -> bool:
        """Updates a existing Submodel.

        :param submodel_identifier: Identifier of the submodel to update
        :param submodel: Submodel data to update
        :return: True if the update was successful, False otherwise
        """
        if not self._client.submodels:
            _logger.error("Submodel API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return False

        sm_data = _to_dict(submodel)

        if sm_data is None:
            return False

        return self._client.submodels.put_submodels_by_id(submodel_identifier, sm_data)

    # DELETE /submodels/{submodelIdentifier}
    def delete_submodel_by_id(self, submodel_identifier: str) -> bool:
        """Deletes a Submodel.

        :param submodel_identifier: ID of the submodel to delete
        :return: True if the deletion was successful, False otherwise
        """
        if not self._client.submodels:
            _logger.error("Submodel API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return False

        return self._client.submodels.delete_submodel_by_id(submodel_identifier)

    # GET /submodels/{submodelIdentifier}/submodel-elements/{idShortPath}
    def get_submodel_element_by_path_submodel_repo(
        self, submodel_identifier: str, id_short_path: str, level: Level = Level.default, extent: Extent = Extent.default
    ) -> model.SubmodelElement | None:
        """Returns a specific submodel element from the Submodel at a specified path.

        :param submodel_identifier: Encoded ID of the Submodel to retrieve element from
        :param id_short_path: Path of the Submodel element to retrieve
        :param level: Determines the structural depth of the respective resource content. Available values : deep, core
        :param extent: Determines to which extent the resource is being serialized. Available values : withBlobValue, withoutBlobValue
        :return: Submodel element data or None if an error occurred
        """
        if not self._client.submodels:
            _logger.error("Submodel API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return None

        content = self._client.submodels.get_submodel_element_by_path_submodel_repo(submodel_identifier, id_short_path, str(level), str(extent))

        if not content:
            return None

        return _to_object(content)

    # PUT /submodels/{submodelIdentifier}/submodel-elements/{idShortPath}
    def put_submodel_element_by_path_submodel_repo(
        self, submodel_identifier: str, id_short_path: str, submodel_element: model.SubmodelElement, level: Level = Level.default
    ) -> bool:
        """Updates a submodel element at a specified path within the submodel elements hierarchy.

        :param submodel_identifier: Encoded ID of the Submodel to update element for
        :param id_short_path: Path of the Submodel element to update
        :param request_body: Submodel element data to update as dictionary
        :param level: Determines the structural depth of the respective resource content. Available values : deep, core
        :return: True if the update was successful, False otherwise
        """
        if not self._client.submodels:
            _logger.error("Submodel API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return False

        sme_data = _to_dict(submodel_element)

        if sme_data is None:
            return False

        return self._client.submodels.put_submodel_element_by_path_submodel_repo(submodel_identifier, id_short_path, sme_data, str(level))

    # POST /submodels/{submodelIdentifier}/submodel-elements/{idShortPath}
    def post_submodel_element_by_path_submodel_repo(
        self,
        submodel_identifier: str,
        id_short_path: str,
        submodel_element: model.SubmodelElement,
        level: Level = Level.default,
        extent: Extent = Extent.default,
    ) -> model.SubmodelElement | None:
        """Creates a new submodel element at a specified path within submodel elements hierarchy.

        :param submodel_identifier: Encoded ID of the submodel to create elements for
        :param id_short_path: Path within the Submodel elements hierarchy
        :param submodel_element: The new Submodel element
        :param level: Determines the structural depth of the respective resource content. Available values : deep, core
        :param extent: Determines to which extent the resource is being serialized. Available values : withBlobValue, withoutBlobValue
        :return: Submodel element object or None if an error occurred
        """
        if not self._client.submodels:
            _logger.error("Submodel API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return None

        sme_data = _to_dict(submodel_element)

        if sme_data is None:
            return None

        content = self._client.submodels.post_submodel_element_by_path_submodel_repo(
            submodel_identifier, id_short_path, sme_data, str(level), str(extent)
        )

        if not content:
            return None

        return _to_object(content)

    # DELETE /submodels/{submodelIdentifier}/submodel-elements/{idShortPath}
    # TODO write test
    def delete_submodel_element_by_path_submodel_repo(self, submodel_identifier: str, id_short_path: str) -> bool:
        """Deletes a submodel element at a specified path within the submodel elements hierarchy.

        :param submodel_identifier: Encoded ID of the Submodel to delete submodel element from
        :param id_short_path: Path of the Submodel element to delete
        :return: True if the deletion was successful, False otherwise
        """
        if not self._client.submodels:
            _logger.error("Submodel API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return False

        return self._client.submodels.delete_submodel_element_by_path_submodel_repo(submodel_identifier, id_short_path)

    # GET /submodels
    def get_all_submodels(
        self,
        semantic_id: str = "",
        id_short: str = "",
        limit: int = 0,
        cursor: str = "",
        level: Level = Level.default,
        extent: Extent = Extent.default,
    ) -> SubmodelPaginatedData | None:
        """Returns all Submodels.

        :param semantic_id: The value of the semantic id reference (UTF8-BASE64-URL-encoded)
        :param id_short: The idShort of the Submodel
        :param limit: Maximum number of Submodels to return
        :param cursor: Cursor for pagination
        :param level: Determines the structural depth of the respective resource content. Available values : deep, core
        :param extent: Determines to which extent the resource is being serialized. Available values : withBlobValue, withoutBlobValue
        :return: List of Submodel or None if an error occurred
        """
        if not self._client.submodels:
            _logger.error("Submodel API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return None

        content = self._client.submodels.get_all_submodels(semantic_id, id_short, limit, cursor, str(level), str(extent))

        if not content:
            return None

        return create_submodel_paging_data(content)

    # POST /submodels
    def post_submodel(self, submodel: model.Submodel) -> model.Submodel | None:
        """Creates a new Submodel.

        :param submodel: Submodel to post
        :return: Submodel or None if an error occurred
        """
        if not self._client.submodels:
            _logger.error("Submodel API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return None

        sm_data = _to_dict(submodel)

        if sm_data is None:
            return None

        content = self._client.submodels.post_submodel(sm_data)

        if not content:
            return None

        return _to_object(content)

    # GET /submodels/{submodelIdentifier}/submodel-elements
    def get_all_submodel_elements_submodel_repository(
        self,
        submodel_identifier: str,
    ) -> SubmodelElementPaginatedData | None:
        """Returns all submodel elements including their hierarchy. !!!Serialization to model.SubmodelElement currently not possible.

        :param submodel_identifier: Encoded ID of the Submodel to retrieve elements from
        :return: List of Submodel elements or None if an error occurred
        """
        if not self._client.submodels:
            _logger.error("Submodel API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return None

        content = self._client.submodels.get_all_submodel_elements_submodel_repository(submodel_identifier)

        if not content:
            return None

        return create_submodel_element_paging_data(content)

    # POST /submodels/{submodelIdentifier}/submodel-elements
    def post_submodel_element_submodel_repo(self, submodel_identifier: str, submodel_element: model.SubmodelElement) -> model.SubmodelElement | None:
        """Creates a new submodel element.

        :param submodel_identifier: Encoded ID of the Submodel to create elements for
        :param request_body: Submodel element
        :return: Submodel or None if an error occurred
        """
        if not self._client.submodels:
            _logger.error("Submodel API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return None

        sme_data = _to_dict(submodel_element)

        if sme_data is None:
            return None

        content = self._client.submodels.post_submodel_element_submodel_repo(submodel_identifier, sme_data)

        if not content:
            return None

        return _to_object(content)

    # POST /submodels/{submodelIdentifier}/submodel-elements/{idShortPath}/invoke
    def invoke_operation_submodel_repo(self, submodel_identifier: str, id_short_path: str, request_body: dict, async_: str = "async") -> dict | None:
        """Synchronously invokes an Operation at a specified path.

        :param submodel_identifier: The Submodels unique id
        :param id_short_path: IdShort path to the operation element (dot-separated)
        :param request_body: Input parameters for the operation
        :param async_: Determines whether an operation invocation is performed asynchronously or synchronously
        :return: Operation result or None if an error occurred
        """
        if not self._client.submodels:
            _logger.error("Submodel API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return None

        content = self._client.submodels.invoke_operation_submodel_repo(submodel_identifier, id_short_path, request_body, async_)

        if not content:
            return None

        return content

    def get_submodel_element_by_path_value_only_submodel_repo(self, submodel_identifier: str, id_short_path: str) -> str | None:
        """Retrieves the value of a specific SubmodelElement.

        :param submodel_identifier: The Submodels unique id
        :param id_short_path: IdShort path to the submodel element (dot-separated)
        :return: Submodel element value or None if an error occurred
        """
        if not self._client.submodels:
            _logger.error("Submodel API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return None

        return self._client.submodels.get_submodel_element_by_path_value_only_submodel_repo(submodel_identifier, id_short_path)

    # PATCH /submodels/{submodelIdentifier}/submodel-elements/{idShortPath}/$value
    def patch_submodel_element_by_path_value_only_submodel_repo(self, submodel_identifier: str, submodel_element_path: str, value: str) -> bool:
        """Updates the value of an existing SubmodelElement.

        :param submodel_identifier: Encoded ID of the Submodel to update submodel element for
        :param submodel_element_path: Path of the Submodel element to update
        :param value: Submodel element value to update as string
        :return: True if the patch was successful, False otherwise
        """
        if not self._client.submodels:
            _logger.error("Submodel API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return False

        return self._client.submodels.patch_submodel_element_by_path_value_only_submodel_repo(submodel_identifier, submodel_element_path, value)

    # GET /submodels/{submodelIdentifier}/$value
    def get_submodel_by_id_value_only(self, submodel_identifier: str, level: Level = Level.default, extent: Extent = Extent.default) -> dict | None:
        """Returns the value of a specific Submodel.

        :param submodel_identifier: Encoded ID of the Submodel to retrieve
        :param level: Determines the structural depth of the respective resource content. Available values : deep, core
        :param extent: Determines to which extent the resource is being serialized. Available values : withBlobValue, withoutBlobValue
        :return: Submodel value as dictionary or None if an error occurred
        """
        if not self._client.submodels:
            _logger.error("Submodel API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return None

        content = self._client.submodels.get_submodel_by_id_value_only(submodel_identifier, str(level), str(extent))

        if not content:
            return None

        return content

    # PATCH /submodels/{submodelIdentifier}/$value
    def patch_submodel_by_id_value_only(self, submodel_identifier: str, request_body: dict, level: Level = Level.default) -> bool:
        """Updates the values of an existing Submodel.

        :param submodel_identifier: The Submodels unique id
        :param request_body: Submodel values to update as dict
        :param level: Determines the structural depth of the respective resource content. Available values : deep, core
        :return: True if the patch was successful, False otherwise
        """
        if not self._client.submodels:
            _logger.error("Submodel API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return False

        return self._client.submodels.patch_submodel_by_id_value_only(submodel_identifier, request_body, str(level))

    # GET /submodels/{submodelIdentifier}/$metadata
    def get_submodel_by_id_metadata(self, submodel_identifier: str, level: str = "") -> dict | None:
        """Returns the metadata attributes of a specific Submodel.

        :param submodel_identifier: The Submodels unique id
        :param level: Determines the structural depth of the respective resource content. Available values : deep, core
        :return: Metadata attributes of the Submodel as dict or None if an error occurred
        """
        if not self._client.submodels:
            _logger.error("Submodel API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return None

        content = self._client.submodels.get_submodel_by_id_metadata(submodel_identifier, str(level))

        if not content:
            return None

        return content

    # not supported by Java Server

    # PATCH /submodels/{submodelIdentifier}
    def patch_submodel_by_id(self, submodel_identifier: str, submodel: model.Submodel) -> bool:
        """Updates an existing Submodel.

        :param submodel_identifier: Encoded ID of the Submodel to delete
        :return: True if the patch was successful, False otherwise
        """
        if not self._client.submodels:
            _logger.error("Submodel API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return False

        sm_data = _to_dict(submodel)

        if sm_data is None:
            return False

        return self._client.submodels.patch_submodel_by_id(submodel_identifier, sm_data)

    # endregion

    # region shell registry

    # currently no SDK implementation for descriptor classes -> no implementation for wrapper

    # endregion

    # region experimental

    # GET /submodels/{submodelIdentifier}/submodel-elements/{idShortPath}/attachment
    def experimental_get_file_by_path_submodel_repo(self, submodel_identifier: str, id_short_path: str) -> Attachment | None:
        """Downloads file content from a specific submodel element from the Submodel at a specified path. Experimental feature - may not be supported by all servers.

        :param submodel_identifier: The Submodels unique id
        :param id_short_path: IdShort path to the submodel element (dot-separated)
        :return: Attachment object with file content as bytes (octet-stream) or None if an error occurred
        """
        if not self._client.experimental:
            _logger.error("Experimental API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return None

        sme = self.get_submodel_element_by_path_submodel_repo(submodel_identifier, id_short_path)

        if not sme or not isinstance(sme, model.File):
            _logger.warning(f"No submodel element found at path '{id_short_path}' in submodel '{submodel_identifier}' on server.")
            return None

        byte_content = self._client.experimental.get_file_by_path_submodel_repo(submodel_identifier, id_short_path)

        if not byte_content:
            _logger.warning(f"No file found at path '{id_short_path}' in submodel '{submodel_identifier}' on server.")
            return None

        return Attachment(
            content=byte_content,
            content_type=puremagic.from_string(byte_content, mime=True),
            filename=sme.value,
        )

    # POST /submodels/{submodelIdentifier}/submodel-elements/{idShortPath}/attachment
    def experimental_post_file_by_path_submodel_repo(self, submodel_identifier: str, id_short_path: str, file: Path) -> bool:
        """Uploads file content to an existing submodel element at a specified path within submodel elements hierarchy. Experimental feature - may not be supported by all servers.

        :param submodel_identifier: The Submodels unique id
        :param id_short_path: IdShort path to the submodel element (dot-separated)
        :param file: Path to the file to upload as attachment
        :return: Attachment data as bytes or None if an error occurred
        """
        if not self._client.experimental:
            _logger.error("Experimental API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return False

        return self._client.experimental.post_file_by_path_submodel_repo(submodel_identifier, id_short_path, file)

    def experimental_put_file_by_path_submodel_repo(self, submodel_identifier: str, id_short_path: str, file: Path) -> bool:
        """Uploads file content to an existing submodel element at a specified path within submodel elements hierarchy. Experimental feature - may not be supported by all servers.

        :param submodel_identifier: The Submodels unique id
        :param id_short_path: IdShort path to the submodel element (dot-separated)
        :param file: Path to the file to upload as attachment
        :return: Attachment data as bytes or None if an error occurred
        """
        if not self._client.experimental:
            _logger.error("Experimental API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return False

        return self._client.experimental.put_file_by_path_submodel_repo(submodel_identifier, id_short_path, file)

    def experimental_delete_file_by_path_submodel_repo(self, submodel_identifier: str, id_short_path: str) -> bool:
        """Deletes file content of an existing submodel element at a specified path within submodel elements hierarchy. Experimental feature - may not be supported by all servers.

        :param submodel_identifier: The Submodels unique id
        :param id_short_path: IdShort path to the submodel element (dot-separated)
        :return: True if deletion was successful, False otherwise
        """
        if not self._client.experimental:
            _logger.error("Experimental API is not initialized in the client. Call 'initialize()' method of the client before calling this method.")
            return False

        return self._client.experimental.delete_file_by_path_submodel_repo(submodel_identifier, id_short_path)

    # endregion


# region wrapper


def create_by_url(
    base_url: str,
    basic_auth_username: str = "",
    basic_auth_password: str = "",
    o_auth_client_id: str = "",
    o_auth_client_secret: str = "",
    o_auth_token_url: str = "",
    bearer_auth_token: str = "",
    http_proxy: str = "",
    https_proxy: str = "",
    time_out: int = 200,
    connection_time_out: int = 60,
    ssl_verify: bool = True,  # noqa: FBT001, FBT002
    trust_env: bool = True,  # noqa: FBT001, FBT002
    encoded_ids: bool = True,  # noqa: FBT001, FBT002
) -> SdkWrapper | None:
    """Create a wrapper for a AAS server connection from the given parameters.

    :param base_url: Base URL of the AAS server, e.g. "http://basyx_python_server:80/"
    :param basic_auth_username: Username for the AAS server basic authentication, defaults to ""
    :param basic_auth_password: Password for the AAS server basic authentication, defaults to ""
    :param o_auth_client_id: Client ID for OAuth authentication, defaults to ""
    :param o_auth_client_secret: Client secret for OAuth authentication, defaults to ""
    :param o_auth_token_url: Token URL for OAuth authentication, defaults to ""
    :param bearer_auth_token: Bearer token for authentication, defaults to ""
    :param http_proxy: HTTP proxy URL, defaults to ""
    :param https_proxy: HTTPS proxy URL, defaults to ""
    :param time_out: Timeout for the API calls, defaults to 200
    :param connection_time_out: Timeout for the connection to the API, defaults to 60
    :param ssl_verify: Whether to verify SSL certificates, defaults to True
    :param trust_env: Whether to trust environment variables for proxy settings, defaults to True
    :param encoded_ids: If enabled, all IDs used in API requests have to be base64-encoded
    :return: An instance of SdkWrapper initialized with the provided parameters or None if initialization fails
    """
    _logger.info(f"Create AAS server http client from URL '{base_url}'.")
    config_dict: dict[str, Any] = {}
    config_dict["BaseUrl"] = base_url
    config_dict["HttpProxy"] = http_proxy
    config_dict["HttpsProxy"] = https_proxy
    config_dict["TimeOut"] = str(time_out)
    config_dict["ConnectionTimeOut"] = str(connection_time_out)
    config_dict["SslVerify"] = str(ssl_verify)
    config_dict["TrustEnv"] = str(trust_env)
    config_dict["EncodedIds"] = str(encoded_ids)

    config_dict["AuthenticationSettings"] = {
        "BasicAuth": {"Username": basic_auth_username},
        "OAuth": {
            "ClientId": o_auth_client_id,
            "TokenUrl": o_auth_token_url,
        },
    }

    return create_by_dict(config_dict, basic_auth_password, o_auth_client_secret, bearer_auth_token)


def create_by_dict(
    configuration: dict, basic_auth_password: str = "", o_auth_client_secret: str = "", bearer_auth_token: str = ""
) -> SdkWrapper | None:
    """Create a wrapper for a AAS server connection from the given configuration.

    :param configuration: Dictionary containing the AAS server connection settings
    :param basic_auth_password: Password for the AAS server basic authentication, defaults to ""
    :param o_auth_client_secret: Client secret for OAuth authentication, defaults to ""
    :param bearer_auth_token: Bearer token for authentication, defaults to ""
    :return: An instance of SdkWrapper initialized with the provided parameters or None if initialization fails
    """
    _logger.info("Create AAS server wrapper from dictionary.")
    config_string = json.dumps(configuration, indent=4)
    return SdkWrapper(config_string, basic_auth_password, o_auth_client_secret, bearer_auth_token)


def create_by_config(
    config_file: Path, basic_auth_password: str = "", o_auth_client_secret: str = "", bearer_auth_token: str = ""
) -> SdkWrapper | None:
    """Create a wrapper for a AAS server connection from a given configuration file.

    :param config_file: Path to the configuration file containing the AAS server connection settings
    :param basic_auth_password: Password for the AAS server basic authentication, defaults to ""
    :param o_auth_client_secret: Client secret for OAuth authentication, defaults to ""
    :param bearer_auth_token: Bearer token for authentication, defaults to ""
    :return: An instance of SdkWrapper initialized with the provided parameters or None if initialization fails
    """
    _logger.info(f"Create AAS wrapper client from configuration file '{config_file}'.")
    if not config_file.exists():
        config_string = "{}"
        _logger.warning(f"Configuration file '{config_file}' not found. Using default config.")
    else:
        config_string = config_file.read_text(encoding="utf-8")
        _logger.debug(f"Configuration file '{config_file}' found.")
    return SdkWrapper(config_string, basic_auth_password, o_auth_client_secret, bearer_auth_token)


# endregion



from http import client
import pytest
from pathlib import Path
from aas_http_client.classes.wrapper.sdk_wrapper import IdEncoding, Level, create_by_config, SdkWrapper, create_by_dict, create_by_url
from basyx.aas import model
import aas_http_client.utilities.model_builder as model_builder
import aas_http_client.utilities.sdk_tools as sdk_tools
from urllib.parse import urlparse
import json
from aas_http_client.utilities import encoder
import random

JAVA_SERVER_PORTS = [8075]
PYTHON_SERVER_PORTS = [5080, 80]
DOTNET_SERVER_PORTS = [5043]

AIMC_SM_ID = "https://fluid40.de/ids/sm/7644_4034_2556_2369"
SM_ID = "fluid40/sm_http_client_unit_tests"
SHELL_ID = "fluid40/aas_http_client_unit_tests"

CONFIG_FILES = [
    "./tests/server_configs/test_java_server_config.yml",
    "./tests/server_configs/test_dotnet_server_config.yml",
    "./tests/server_configs/test_python_server_config.yml"
]

# CONFIG_FILES = [
#     "./tests/server_configs/test_dotnet_server_config.yml",
# ]

@pytest.fixture(params=CONFIG_FILES, scope="module")
def wrapper(request) -> SdkWrapper:
    try:
        file = Path(request.param).resolve()

        if not file.exists():
            raise FileNotFoundError(f"Configuration file {file} does not exist.")

        wrapper = create_by_config(file, basic_auth_password="")

        rand = random.randint(0, 10)
        if (rand % 2) == 0:
            wrapper.set_encoded_ids(IdEncoding.encoded)

        wrapper.set_encoded_ids(IdEncoding.encoded)

    except Exception as e:
        raise RuntimeError("Unable to connect to server.")

    shells = wrapper.get_all_asset_administration_shells()
    if shells is None:
        raise RuntimeError("No shells found on server. Please check the server configuration.")

    return wrapper

@pytest.fixture(scope="module")
def shared_sme_string() -> model.Property:
    # create a Submodel
    return model_builder.create_base_submodel_element_property("sme_property_string", model.datatypes.String, "Sample String Value", description="This is a sample string property for unit testing.", display_name="Sample String Property")

@pytest.fixture(scope="module")
def shared_sme_bool() -> model.Property:
    # create a Submodel
    return model_builder.create_base_submodel_element_property("sme_property_bool", model.datatypes.Boolean, True, description="This is a sample boolean property for unit testing.", display_name="Sample Boolean Property"    )

@pytest.fixture(scope="module")
def shared_sme_int() -> model.Property:
    # create a Submodel
    return model_builder.create_base_submodel_element_property("sme_property_int", model.datatypes.Integer, 262, description="This is a sample integer property for unit testing.", display_name="Sample Integer Property")

@pytest.fixture(scope="module")
def shared_sme_float() -> model.Property:
    # create a Submodel
    return model_builder.create_base_submodel_element_property("sme_property_float", model.datatypes.Float, 262.3, description="This is a sample float property for unit testing.", display_name="Sample Float Property")

@pytest.fixture(scope="module")
def shared_sm() -> model.Submodel:
    # create a Submodel
    submodel = model_builder.create_base_submodel(identifier=SM_ID, id_short="sm_http_client_unit_tests", display_name="Submodel HTTP Client Unit Tests", description="This is a sample Submodel created for unit testing of the AAS HTTP Client.")
    submodel.category = "Unit Test"
    return submodel

@pytest.fixture(scope="module")
def shared_aas(shared_sm: model.Submodel) -> model.AssetAdministrationShell:
    # create an AAS
    aas = model_builder.create_base_aas(identifier=SHELL_ID, id_short="aas_http_client_unit_tests", global_asset_identifier=SHELL_ID, display_name="AAS HTTP Client Unit Tests", description="This is a sample AAS created for unit testing of the AAS HTTP Client.")

    # add Submodel to AAS
    sdk_tools.add_submodel_to_aas(aas, shared_sm)

    return aas

def test_000a_create_wrapper_by_url(wrapper: SdkWrapper):
    base_url: str = wrapper.base_url
    new_client: SdkWrapper = create_by_url(base_url=base_url)
    assert new_client is not None

def test_000b_create_wrapper_by_dict(wrapper: SdkWrapper):
    base_url: str = wrapper.base_url

    config_dict: dict = {
        "BaseUrl": base_url
    }

    new_client: SdkWrapper = create_by_dict(configuration=config_dict)
    assert new_client is not None

def test_000c_get_client(wrapper: SdkWrapper):
    client = wrapper.get_client()
    assert client is not None
    root = client.get_root()
    assert root is not None

def test_001a_connect(wrapper: SdkWrapper):
    assert wrapper is not None

def test_001b_delete_all_asset_administration_shells(wrapper: SdkWrapper):
    result = wrapper.get_all_asset_administration_shells()
    assert result is not None

    for shell in result.results:
        shell_id = shell.id

        if wrapper.get_encoded_ids() == IdEncoding.encoded:
            shell_id = encoder.encode_base_64(shell_id)

        if shell_id:
            delete_result = wrapper.delete_asset_administration_shell_by_id(shell_id)
            assert delete_result

    shells_result = wrapper.get_all_asset_administration_shells()
    assert len(shells_result.results) == 0

def test_001c_delete_all_submodels(wrapper: SdkWrapper):
    result = wrapper.get_all_submodels()
    assert result is not None

    for submodel in result.results:
        submodel_id = submodel.id

        if wrapper.get_encoded_ids() == IdEncoding.encoded:
            submodel_id = encoder.encode_base_64(submodel_id)

        if submodel_id:
            delete_result = wrapper.delete_submodel_by_id(submodel_id)
            assert delete_result

    submodels_result = wrapper.get_all_submodels()
    assert len(submodels_result.results) == 0

def test_002_get_all_asset_administration_shells(wrapper: SdkWrapper):
    shells = wrapper.get_all_asset_administration_shells()
    assert shells is not None
    assert len(shells.results) == 0

def test_003_post_asset_administration_shell(wrapper: SdkWrapper, shared_aas: model.AssetAdministrationShell):
    shell = wrapper.post_asset_administration_shell(shared_aas)

    assert shell is not None
    assert shell.id == SHELL_ID
    assert shell.id_short == shared_aas.id_short

    shells = wrapper.get_all_asset_administration_shells()
    assert shells is not None
    assert len(shells.results) == 1
    assert shells.results[0].id_short == shared_aas.id_short
    assert shells.results[0].id == SHELL_ID

def test_004a_get_asset_administration_shell_by_id(wrapper: SdkWrapper, shared_aas: model.AssetAdministrationShell):
    shell_id = SHELL_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        shell_id = encoder.encode_base_64(shell_id)

    shell = wrapper.get_asset_administration_shell_by_id(shell_id)

    assert shell is not None
    assert shell.id_short == shared_aas.id_short
    assert shell.id == SHELL_ID

def test_004b_get_asset_administration_shell_by_id(wrapper: SdkWrapper):
    shell = wrapper.get_asset_administration_shell_by_id("non_existent_id")

    assert shell is None

def test_005a_put_asset_administration_shell_by_id(wrapper: SdkWrapper, shared_aas: model.AssetAdministrationShell):
    aas = model.AssetAdministrationShell(id_=shared_aas.asset_information.global_asset_id, asset_information=shared_aas.asset_information)
    aas.id_short = shared_aas.id_short

    description_text = "Put description for unit tests"
    aas.description = model.MultiLanguageTextType({"en": description_text})
    aas.submodel = shared_aas.submodel  # Keep existing submodels

    shell_id = SHELL_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        shell_id = encoder.encode_base_64(shell_id)

    result = wrapper.put_asset_administration_shell_by_id(shell_id, aas)

    assert result

    shell = wrapper.get_asset_administration_shell_by_id(shell_id)

    assert shell is not None
    assert shell.id_short == shared_aas.id_short
    assert shell.id == SHELL_ID
    # description must have changed
    assert shell.description.get("en", "") == description_text
    assert shell.description.get("en", "") != shared_aas.description.get("en", "")
    # submodels must be retained
    assert len(shell.submodel) == len(shared_aas.submodel)

    # The display name must be empty
    # currently not working in dotnet
    # assert len(get_result.get("displayName", {})) == 0

    # # restore to its original state
    wrapper.put_asset_administration_shell_by_id(shell_id, shared_aas)  # Restore original submodel

def test_005b_put_asset_administration_shell_by_id(wrapper: SdkWrapper, shared_aas: model.AssetAdministrationShell):
    # put with other ID
    id_short = "put_short_id"
    identifier = f"fluid40/{id_short}"
    asset_info = model_builder.create_base_asset_information(identifier)
    aas = model.AssetAdministrationShell(id_=asset_info.global_asset_id, asset_information=asset_info)
    aas.id_short = id_short

    description_text = {"en": "Updated description for unit tests"}
    aas.description = model.MultiLanguageTextType(description_text)

    shell_id = SHELL_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        shell_id = encoder.encode_base_64(SHELL_ID)

    parsed = urlparse(wrapper.base_url)
    if int(parsed.port) in PYTHON_SERVER_PORTS:
        # NOTE: Python server crashes by this test
        result = False
    else:
        result = wrapper.put_asset_administration_shell_by_id(shell_id, aas)

    assert not result

    shell = wrapper.get_asset_administration_shell_by_id(shell_id)
    assert shell.description.get("en", "") != description_text
    assert shell.description.get("en", "") == shared_aas.description.get("en", "")

def test_006_get_asset_administration_shell_by_id_reference_aas_repository(wrapper: SdkWrapper):
    shell_id = SHELL_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        shell_id = encoder.encode_base_64(SHELL_ID)

    reference = wrapper.get_asset_administration_shell_by_id_reference_aas_repository(shell_id)

    assert reference is not None
    assert len(reference.key) == 1
    assert reference.key[0].value == SHELL_ID

def test_007_get_submodel_by_id_aas_repository(wrapper: SdkWrapper, shared_sm: model.Submodel):
    shell_id = SHELL_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        shell_id = encoder.encode_base_64(SHELL_ID)

    submodel = wrapper.get_submodel_by_id_aas_repository(shell_id, shared_sm.id)

    assert submodel is None

def test_008_get_all_submodels(wrapper: SdkWrapper):
    submodels = wrapper.get_all_submodels()
    assert submodels is not None
    assert len(submodels.results) == 0

def test_009a_post_submodel(wrapper: SdkWrapper, shared_sm: model.Submodel):
    submodel = wrapper.post_submodel(shared_sm)

    assert submodel is not None
    assert submodel.id == SM_ID
    assert submodel.id_short == shared_sm.id_short

    submodels = wrapper.get_all_submodels()
    assert submodels is not None
    assert len(submodels.results) == 1
    assert submodels.results[0].id_short == shared_sm.id_short
    assert submodels.results[0].id == SM_ID

def test_009b_post_submodel(wrapper: SdkWrapper):
    sm_template_file = Path(f"./tests/test_data/aimc.json").resolve()

    with Path.open(sm_template_file, "r", encoding="utf-8") as f:
        sm_data = json.load(f)

    submodel = sdk_tools.convert_to_object(sm_data)

    result = wrapper.post_submodel(submodel)

    assert result is not None
    assert result.id == AIMC_SM_ID

    get_result = wrapper.get_all_submodels()
    assert get_result is not None
    submodels = get_result.results
    assert len(submodels) == 2

def test_010_get_submodel_by_id_aas_repository(wrapper: SdkWrapper, shared_sm: model.Submodel):
    shell_id = SHELL_ID
    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        shell_id = encoder.encode_base_64(SHELL_ID)
        sm_id = encoder.encode_base_64(SM_ID)

    submodel = wrapper.get_submodel_by_id_aas_repository(shell_id, sm_id)

    parsed = urlparse(wrapper.base_url)
    if int(parsed.port) in JAVA_SERVER_PORTS:
        # Basyx java server do not provide this endpoint
        assert submodel is None
    else:
        assert submodel is not None
        assert submodel.id_short == shared_sm.id_short
        assert submodel.id == SM_ID

def test_011a_get_submodel_by_id(wrapper: SdkWrapper, shared_sm: model.Submodel):
    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    submodel = wrapper.get_submodel_by_id(sm_id)

    assert submodel is not None
    assert submodel.id_short == shared_sm.id_short
    assert submodel.id == SM_ID

def test_011b_get_submodel_by_id(wrapper: SdkWrapper):
    result = wrapper.get_submodel_by_id("non_existent_id")

    assert result is None

def test_011c_get_submodel_by_id(wrapper: SdkWrapper):
    sm_id = AIMC_SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(AIMC_SM_ID)

    result = wrapper.get_submodel_by_id(sm_id)

    assert result is not None
    assert result.id == AIMC_SM_ID

def test_011d_get_submodel_by_id(wrapper: SdkWrapper):
    sm_id = AIMC_SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(AIMC_SM_ID)

    result = wrapper.get_submodel_by_id(sm_id, level=Level.core)

    assert result is not None
    assert result.id == AIMC_SM_ID
    # assert "submodelElements" not in result

def test_012_patch_submodel_by_id(wrapper: SdkWrapper, shared_sm: model.Submodel):
    sm = model.Submodel(shared_sm.id_short)
    sm.id_short = shared_sm.id_short

    description_text = "Patched description for unit tests"
    sm.description = model.MultiLanguageTextType({"en": description_text})

    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    result = wrapper.patch_submodel_by_id(sm_id, sm)

    parsed = urlparse(wrapper.base_url)
    if int(parsed.port) in JAVA_SERVER_PORTS or int(parsed.port) in PYTHON_SERVER_PORTS:
        # Basyx java server do not provide this endpoint
        assert not result
    else:
        assert result

        submodel = wrapper.get_submodel_by_id(sm_id)
        assert submodel is not None
        assert submodel.id_short == shared_sm.id_short
        assert submodel.id == SM_ID
        # Only the description may change in patch.
        assert submodel.description.get("en", "") == description_text
        assert submodel.description.get("en", "") != shared_sm.description.get("en", "")
        # The display name must remain the same.
        assert submodel.display_name == shared_sm.display_name
        assert len(submodel.submodel_element) == len(shared_sm.submodel_element)

def test_013_put_submodel_by_id_aas_repository(wrapper: SdkWrapper, shared_sm: model.Submodel):
    sm = model.Submodel(SM_ID)
    sm.id_short = shared_sm.id_short

    description_text = "Put via shell description for unit tests"
    sm.description = model.MultiLanguageTextType({"en": description_text})
    sm.display_name = shared_sm.display_name  # Keep existing display name because of problems with empty lists

    shell_id = SHELL_ID
    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        shell_id = encoder.encode_base_64(SHELL_ID)
        sm_id = encoder.encode_base_64(SM_ID)

    result = wrapper.put_submodel_by_id_aas_repository(shell_id, sm_id, sm)

    parsed = urlparse(wrapper.base_url)
    if int(parsed.port) in JAVA_SERVER_PORTS:
        # Basyx java server do not provide this endpoint
        assert not result
    else:
        assert result

        submodel = wrapper.get_submodel_by_id_aas_repository(shell_id, sm_id)
        assert submodel is not None
        assert submodel.id_short == shared_sm.id_short
        assert submodel.id == SM_ID
        # description must have changed
        assert submodel.description.get("en", "") == description_text
        assert submodel.description.get("en", "") != shared_sm.description.get("en", "")
        # display name stays
        assert submodel.display_name == shared_sm.display_name
        # category was not set an must be empty
        assert submodel.category is None
        assert len(submodel.submodel_element) == 0

        # restore to its original state
        wrapper.put_submodel_by_id_aas_repository(shell_id, sm_id, shared_sm)  # Restore original submodel

def test_014_put_submodels_by_id(wrapper: SdkWrapper, shared_sm: model.Submodel):
    sm = model.Submodel(shared_sm.id)
    sm.id_short = shared_sm.id_short

    description_text = "Put description for unit tests"
    sm.description = model.MultiLanguageTextType({"en": description_text})
    sm.display_name = shared_sm.display_name  # Keep existing display name because of problems with empty lists

    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    result = wrapper.put_submodels_by_id(sm_id, sm)

    assert result

    submodel = wrapper.get_submodel_by_id(sm_id)
    assert submodel is not None
    assert submodel.id_short == shared_sm.id_short
    assert submodel.id == SM_ID
    # description must have changed
    assert submodel.description.get("en", "") == description_text
    assert submodel.description.get("en", "") != shared_sm.description.get("en", "")
    # display name stays
    # assert submodel.display_name == shared_sm.display_name
    # category was not set an must be empty
    assert submodel.category is None
    assert len(submodel.submodel_element) == 0

    # restore to its original state
    wrapper.put_submodels_by_id(sm_id, shared_sm)  # Restore original submodel

def test_015_get_all_submodel_elements_submodel_repository(wrapper: SdkWrapper):
    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    submodel_elements = wrapper.get_all_submodel_elements_submodel_repository(sm_id)

    assert submodel_elements is not None
    assert len(submodel_elements.results) == 0

def test_016a_post_submodel_element_submodel_repo(wrapper: SdkWrapper, shared_sme_string: model.Property):
    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    submodel_element = wrapper.post_submodel_element_submodel_repo(sm_id, shared_sme_string)

    assert submodel_element is not None

    assert isinstance(submodel_element, model.Property)
    property: model.Property = submodel_element
    assert property.value == shared_sme_string.value

    assert submodel_element.id_short == shared_sme_string.id_short
    assert submodel_element.description.get("en", "") == shared_sme_string.description.get("en", "")
    assert submodel_element.display_name.get("en", "") == shared_sme_string.display_name.get("en", "")
    assert submodel_element.value == shared_sme_string.value

    submodel_elements = wrapper.get_all_submodel_elements_submodel_repository(sm_id)

    assert submodel_elements is not None
    assert len(submodel_elements.results) == 1

def test_016b_post_submodel_element_submodel_repo(wrapper: SdkWrapper, shared_sme_bool: model.Property):
    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    submodel_element = wrapper.post_submodel_element_submodel_repo(sm_id, shared_sme_bool)

    assert submodel_element is not None

    assert isinstance(submodel_element, model.Property)
    property: model.Property = submodel_element
    assert property.value == shared_sme_bool.value

    assert submodel_element.id_short == shared_sme_bool.id_short
    assert submodel_element.description.get("en", "") == shared_sme_bool.description.get("en", "")
    assert submodel_element.display_name.get("en", "") == shared_sme_bool.display_name.get("en", "")
    assert submodel_element.value == shared_sme_bool.value

    submodel_elements = wrapper.get_all_submodel_elements_submodel_repository(sm_id)

    assert submodel_elements is not None
    assert len(submodel_elements.results) == 2

def test_016c_post_submodel_element_submodel_repo(wrapper: SdkWrapper, shared_sme_int: model.Property):
    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    submodel_element = wrapper.post_submodel_element_submodel_repo(sm_id, shared_sme_int)

    assert submodel_element is not None

    assert isinstance(submodel_element, model.Property)
    property: model.Property = submodel_element
    assert property.value == shared_sme_int.value

    assert submodel_element.id_short == shared_sme_int.id_short
    assert submodel_element.description.get("en", "") == shared_sme_int.description.get("en", "")
    assert submodel_element.display_name.get("en", "") == shared_sme_int.display_name.get("en", "")
    assert submodel_element.value == shared_sme_int.value

    submodel_elements = wrapper.get_all_submodel_elements_submodel_repository(sm_id)

    assert submodel_elements is not None
    assert len(submodel_elements.results) == 3

def test_016d_post_submodel_element_submodel_repo(wrapper: SdkWrapper, shared_sme_float: model.Property):
    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    submodel_element = wrapper.post_submodel_element_submodel_repo(sm_id, shared_sme_float)

    assert submodel_element is not None

    assert isinstance(submodel_element, model.Property)
    property: model.Property = submodel_element
    assert property.value == shared_sme_float.value

    assert submodel_element.id_short == shared_sme_float.id_short
    assert submodel_element.description.get("en", "") == shared_sme_float.description.get("en", "")
    assert submodel_element.display_name.get("en", "") == shared_sme_float.display_name.get("en", "")
    assert submodel_element.value == shared_sme_float.value

    submodel_elements = wrapper.get_all_submodel_elements_submodel_repository(sm_id)

    assert submodel_elements is not None
    assert len(submodel_elements.results) == 4

def test_017a_get_submodel_element_by_path_submodel_repo(wrapper: SdkWrapper, shared_sme_string: model.Property):
    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    submodel_element = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_string.id_short)

    assert submodel_element is not None

    assert isinstance(submodel_element, model.Property)

    assert submodel_element.id_short == shared_sme_string.id_short
    assert submodel_element.description.get("en", "") == shared_sme_string.description.get("en", "")
    assert submodel_element.display_name.get("en", "") == shared_sme_string.display_name.get("en", "")
    assert submodel_element.value == shared_sme_string.value

def test_017b_get_submodel_element_by_path_submodel_repo(wrapper: SdkWrapper, shared_sme_bool: model.Property):
    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    submodel_element = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_bool.id_short)

    assert submodel_element is not None

    assert isinstance(submodel_element, model.Property)

    assert submodel_element.id_short == shared_sme_bool.id_short
    assert submodel_element.description.get("en", "") == shared_sme_bool.description.get("en", "")
    assert submodel_element.display_name.get("en", "") == shared_sme_bool.display_name.get("en", "")
    assert submodel_element.value == shared_sme_bool.value

def test_017c_get_submodel_element_by_path_submodel_repo(wrapper: SdkWrapper, shared_sme_int: model.Property):
    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    submodel_element = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_int.id_short)

    assert submodel_element is not None

    assert isinstance(submodel_element, model.Property)

    assert submodel_element.id_short == shared_sme_int.id_short
    assert submodel_element.description.get("en", "") == shared_sme_int.description.get("en", "")
    assert submodel_element.display_name.get("en", "") == shared_sme_int.display_name.get("en", "")
    assert submodel_element.value == shared_sme_int.value

def test_017d_get_submodel_element_by_path_submodel_repo(wrapper: SdkWrapper, shared_sme_float: model.Property):
    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    submodel_element = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_float.id_short)

    assert submodel_element is not None

    assert isinstance(submodel_element, model.Property)

    assert submodel_element.id_short == shared_sme_float.id_short
    assert submodel_element.description.get("en", "") == shared_sme_float.description.get("en", "")
    assert submodel_element.display_name.get("en", "") == shared_sme_float.display_name.get("en", "")
    assert submodel_element.value == shared_sme_float.value

def test_018a_patch_submodel_element_by_path_value_only_submodel_repo(wrapper: SdkWrapper, shared_sme_string: model.Property):
    new_value = "Patched String Value"

    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    submodel_element: model.Property = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_string.id_short)
    old_value = submodel_element.value

    result = wrapper.patch_submodel_element_by_path_value_only_submodel_repo(sm_id, shared_sme_string.id_short, new_value)

    parsed = urlparse(wrapper.base_url)
    if int(parsed.port) in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        assert result is False
    else:
        assert result is True

        submodel_element = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_string.id_short)

        assert submodel_element is not None
        assert submodel_element.id_short == shared_sme_string.id_short
        assert submodel_element.description.get("en", "") == shared_sme_string.description.get("en", "")
        assert submodel_element.display_name.get("en", "") == shared_sme_string.display_name.get("en", "")

        assert isinstance(submodel_element, model.Property)
        property: model.Property = submodel_element
        assert property.value == new_value
        assert property.value != old_value

def test_018b_patch_submodel_element_by_path_value_only_submodel_repo(wrapper: SdkWrapper, shared_sme_bool: model.Property):
    new_value = "false"

    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    submodel_element: model.Property = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_bool.id_short)
    old_value = submodel_element.value

    result = wrapper.patch_submodel_element_by_path_value_only_submodel_repo(sm_id, shared_sme_bool.id_short, new_value)

    parsed = urlparse(wrapper.base_url)
    if int(parsed.port) in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        assert result is False
    else:
        assert result is True

        submodel_element = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_bool.id_short)

        assert submodel_element is not None
        assert submodel_element.id_short == shared_sme_bool.id_short
        assert submodel_element.description.get("en", "") == shared_sme_bool.description.get("en", "")
        assert submodel_element.display_name.get("en", "") == shared_sme_bool.display_name.get("en", "")

        assert isinstance(submodel_element, model.Property)
        property: model.Property = submodel_element
        assert property.value == json.loads(new_value)
        assert property.value != old_value

def test_018c_patch_submodel_element_by_path_value_only_submodel_repo(wrapper: SdkWrapper, shared_sme_int: model.Property):
    new_value = "263"

    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    submodel_element: model.Property = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_int.id_short)
    old_value = submodel_element.value

    result = wrapper.patch_submodel_element_by_path_value_only_submodel_repo(sm_id, shared_sme_int.id_short, new_value)

    parsed = urlparse(wrapper.base_url)
    if int(parsed.port) in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        assert result is False
    else:
        assert result is True

        submodel_element = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_int.id_short)

        assert submodel_element is not None
        assert submodel_element.id_short == shared_sme_int.id_short
        assert submodel_element.description.get("en", "") == shared_sme_int.description.get("en", "")
        assert submodel_element.display_name.get("en", "") == shared_sme_int.display_name.get("en", "")

        assert isinstance(submodel_element, model.Property)
        property: model.Property = submodel_element
        assert property.value == int(new_value)
        assert property.value != old_value

def test_018d_patch_submodel_element_by_path_value_only_submodel_repo(wrapper: SdkWrapper, shared_sme_float: model.Property):
    new_value = "262.1"

    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    submodel_element: model.Property = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_float.id_short)
    old_value = submodel_element.value

    result = wrapper.patch_submodel_element_by_path_value_only_submodel_repo(sm_id, shared_sme_float.id_short, new_value)

    parsed = urlparse(wrapper.base_url)
    if int(parsed.port) in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        assert result is False
    else:
        assert result is True

        submodel_element = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_float.id_short)

        assert submodel_element is not None
        assert submodel_element.id_short == shared_sme_float.id_short
        assert submodel_element.description.get("en", "") == shared_sme_float.description.get("en", "")
        assert submodel_element.display_name.get("en", "") == shared_sme_float.display_name.get("en", "")

        assert isinstance(submodel_element, model.Property)
        property: model.Property = submodel_element
        assert property.value == float(new_value)
        assert property.value != old_value
        assert property.value == 262.1

def test_019a_post_submodel_element_by_path_submodel_repo(wrapper: SdkWrapper):
    submodel_element_list = model.SubmodelElementList(id_short="sme_list_1", type_value_list_element=model.Property, value_type_list_element=model.datatypes.String)

    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    post_list_element_result = wrapper.post_submodel_element_submodel_repo(sm_id, submodel_element_list)

    assert post_list_element_result is not None

    property = model_builder.create_base_submodel_element_property(None, model.datatypes.String, "Value in List")# idShort must be empty for list elements

    result = wrapper.post_submodel_element_by_path_submodel_repo(sm_id, submodel_element_list.id_short, property)

    assert result is not None
    assert result.id_short == property.id_short

    submodel = wrapper.get_submodel_by_id(sm_id)

    assert submodel is not None
    elements = submodel.submodel_element
    assert len(elements) == 5  # 4 previous properties + 1 list
    element = list(elements)[4]
    assert element is not None
    assert isinstance(element, model.SubmodelElementList)

    assert element.id_short == submodel_element_list.id_short
    list_elements = element.value
    assert len(list_elements) == 1
    list_element = list(list_elements)[0]
    assert isinstance(list_element, model.Property)

    assert "hack" in list_element.id_short
    assert list_element.value == property.value

def test_019b_post_submodel_element_by_path_submodel_repo(wrapper: SdkWrapper):
    submodel_element_collection = model.SubmodelElementCollection(id_short="sme_collection_1")

    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    first_result = wrapper.post_submodel_element_submodel_repo(sm_id, submodel_element_collection)

    assert first_result is not None

    property = model_builder.create_base_submodel_element_property("sme_property_in_collection", model.datatypes.String, "Value in List")
    result = wrapper.post_submodel_element_by_path_submodel_repo(sm_id, submodel_element_collection.id_short, property)

    assert result is not None
    assert result.id_short== property.id_short

    submodel = wrapper.get_submodel_by_id(sm_id)

    assert submodel is not None
    elements = submodel.submodel_element
    assert len(elements) == 6
    assert list(elements)[5].id_short == submodel_element_collection.id_short
    list_elements = list(elements)[5].value
    assert len(list_elements) == 1
    assert list(list_elements)[0].id_short == property.id_short
    assert list(list_elements)[0].value == property.value

    base_url: str = wrapper.base_url
    new_wrapper = create_by_url(base_url=base_url)
    assert new_wrapper is not None

    sm = new_wrapper.get_submodel_by_id(AIMC_SM_ID)
    assert sm is None

    decoded_id = encoder.encode_base_64(AIMC_SM_ID)
    decoded_sm = new_wrapper.get_submodel_by_id(decoded_id)
    assert decoded_sm is not None
    assert decoded_sm.id == AIMC_SM_ID

def test_020a_encoded_ids(wrapper: SdkWrapper):
    base_url: str = wrapper.base_url
    new_wrapper: SdkWrapper = create_by_url(base_url=base_url)
    assert new_wrapper is not None

    sm = new_wrapper.get_submodel_by_id(AIMC_SM_ID)
    assert sm is None

    encoded_id = encoder.encode_base_64(AIMC_SM_ID)
    encoded_sm = new_wrapper.get_submodel_by_id(encoded_id)
    assert encoded_sm is not None
    assert encoded_sm.id == AIMC_SM_ID

def test_020b_encoded_ids(wrapper: SdkWrapper):
    base_url: str = wrapper.base_url
    new_wrapper: SdkWrapper = create_by_url(base_url=base_url)
    assert new_wrapper is not None

    sm = new_wrapper.get_asset_administration_shell_by_id(SHELL_ID)
    assert sm is None

    encoded_id = encoder.encode_base_64(SHELL_ID)
    encoded_sm = new_wrapper.get_asset_administration_shell_by_id(encoded_id)
    assert encoded_sm is not None
    assert encoded_sm.id == SHELL_ID

def test_021_post_file_by_path_submodel_repo(wrapper: SdkWrapper):
    parsed = urlparse(wrapper.base_url)
    if int(parsed.port) in JAVA_SERVER_PORTS or int(parsed.port) in PYTHON_SERVER_PORTS:
        # NOTE: python server implementation differs
        # NOTE: Basyx java server do not provide this endpoint
        return

    file_sme = model.File("file_sme", content_type="application/pdf")

    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    file_post_result = wrapper.post_submodel_element_submodel_repo(sm_id, file_sme)
    assert file_post_result is not None

    filename = "https.pdf"
    file = Path(f"./tests/test_data/{filename}").resolve()
    result = wrapper.experimental_post_file_by_path_submodel_repo(sm_id, file_sme.id_short, file)
    assert result is True

    result_sme = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, file_sme.id_short)

    assert result_sme is not None
    assert result_sme.id_short == file_sme.id_short

    assert result_sme.content_type == file_sme.content_type
    assert result_sme.value == f"/{filename}"

def test_022_get_file_content_by_path_submodel_repo(wrapper: SdkWrapper):
    parsed = urlparse(wrapper.base_url)
    if int(parsed.port) in JAVA_SERVER_PORTS or int(parsed.port) in PYTHON_SERVER_PORTS:
        # NOTE: python server implementation differs
        # NOTE: Basyx java server do not provide this endpoint
        return

    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    attachment = wrapper.experimental_get_file_by_path_submodel_repo(sm_id, "file_sme")
    assert attachment is not None
    assert attachment.content_type == "application/pdf"
    assert isinstance(attachment.content, bytes)
    assert len(attachment.content) > 0
    assert attachment.content.startswith(b"%PDF-1.7")
    assert attachment.filename == "/https.pdf"

def test_023_put_file_content_by_path_submodel_repo(wrapper: SdkWrapper):
    parsed = urlparse(wrapper.base_url)
    if int(parsed.port) in JAVA_SERVER_PORTS or int(parsed.port) in PYTHON_SERVER_PORTS:
        # NOTE: python server implementation differs
        # NOTE: Basyx java server do not provide this endpoint
        return

    filename = "aimc.json"
    file = Path(f"./tests/test_data/{filename}").resolve()

    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    result = wrapper.experimental_put_file_by_path_submodel_repo(sm_id, "file_sme", file)
    assert result is True

    get_result = wrapper.experimental_get_file_by_path_submodel_repo(sm_id, "file_sme")
    assert get_result is not None
    assert len(get_result.content) > 0
    assert get_result.content.startswith(b"{\n")
    assert get_result.filename == f"/{filename}"
    assert get_result.content_type == "application/json"

    result_sme = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, "file_sme")
    assert result_sme is not None
    assert result_sme.value == f"/{filename}"

def test_024_delete_file_content_by_path_submodel_repo(wrapper: SdkWrapper):
    parsed = urlparse(wrapper.base_url)
    if int(parsed.port) in JAVA_SERVER_PORTS or int(parsed.port) in PYTHON_SERVER_PORTS:
        # NOTE: python server implementation differs
        # NOTE: Basyx java server do not provide this endpoint
        return

    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    result = wrapper.experimental_delete_file_by_path_submodel_repo(sm_id, "file_sme")
    assert result is True

    get_result = wrapper.experimental_get_file_by_path_submodel_repo(sm_id, "file_sme")
    assert get_result is None

    result_sme = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, "file_sme")
    assert result_sme is not None
    assert result_sme.value == None

def test_025_get_thumbnail_aas_repository(wrapper: SdkWrapper):
    parsed = urlparse(wrapper.base_url)
    if int(parsed.port) in PYTHON_SERVER_PORTS:
        # NOTE: python server implementation differs
        return

    shell_id = SHELL_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        shell_id = encoder.encode_base_64(SHELL_ID)

    result = wrapper.get_thumbnail_aas_repository(shell_id)
    assert result is None

def test_026_put_thumbnail_aas_repository(wrapper: SdkWrapper):
    parsed = urlparse(wrapper.base_url)
    if int(parsed.port) in PYTHON_SERVER_PORTS:
        # NOTE: python server implementation differs
        return

    shell_id = SHELL_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        shell_id = encoder.encode_base_64(SHELL_ID)

    filename = "Pen_Machine.png"
    file = Path(f"./tests/test_data/{filename}").resolve()

    result = wrapper.put_thumbnail_aas_repository(shell_id, file.name, file)
    assert result is True

def test_027_get_thumbnail_aas_repository(wrapper: SdkWrapper):
    parsed = urlparse(wrapper.base_url)
    if int(parsed.port) in PYTHON_SERVER_PORTS:
        # NOTE: python server implementation differs
        return

    shell_id = SHELL_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        shell_id = encoder.encode_base_64(SHELL_ID)

    result = wrapper.get_thumbnail_aas_repository(shell_id)
    assert result is not None

    assert len(result.content) > 0
    assert result.content.startswith(b"\x89PNG\r\n\x1a\n")
    assert result.filename == "thumbnail"
    assert result.content_type == "image/png"

def test_028_delete_thumbnail_aas_repository(wrapper: SdkWrapper):
    parsed = urlparse(wrapper.base_url)
    if int(parsed.port) in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        return

    shell_id = SHELL_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        shell_id = encoder.encode_base_64(SHELL_ID)

    result = wrapper.delete_thumbnail_aas_repository(shell_id)
    assert result is True

    get_result = wrapper.get_thumbnail_aas_repository(shell_id)
    assert get_result is None

def test_029_get_all_submodel_references_aas_repository(wrapper: SdkWrapper):
    shell_id = SHELL_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        shell_id = encoder.encode_base_64(SHELL_ID)

    result = wrapper.get_all_submodel_references_aas_repository(shell_id)
    assert result is not None
    references = result.results
    assert len(references) == 1

def test_030_post_submodel_reference_aas_repository(wrapper: SdkWrapper):
    shell_id = SHELL_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        shell_id = encoder.encode_base_64(SHELL_ID)

    id = "temp_sm_id"
    id_short = "TempSM"
    temp_sml_ref = model.ModelReference.from_referable(model_builder.create_base_submodel(identifier=id, id_short=id_short))

    result = wrapper.post_submodel_reference_aas_repository(shell_id, temp_sml_ref)

    assert result is not None
    assert len(result.get("keys", [])) > 0
    key: dict = result.get("keys", [])[0]
    assert key.get("value", "") == id
    assert key.get("type", "") == "Submodel"

    check_result = wrapper.get_all_submodel_references_aas_repository(shell_id)
    assert check_result is not None
    assert len(check_result.results) == 2

def test_031_delete_submodel_reference_by_id_aas_repository(wrapper: SdkWrapper):
    shell_id = SHELL_ID
    sm_id = "temp_sm_id"

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        shell_id = encoder.encode_base_64(SHELL_ID)
        sm_id = encoder.encode_base_64(sm_id)

    result = wrapper.delete_submodel_reference_by_id_aas_repository(shell_id, sm_id)

    assert result is True

    get_result = wrapper.get_all_submodel_references_aas_repository(shell_id)
    assert get_result is not None
    assert len(get_result.results) == 1

def test_032_put_submodel_element_by_path_submodel_repo(wrapper: SdkWrapper, shared_sme_string: model.Property):
    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    sme: model.Property = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_string.id_short)
    old_value = sme.value

    new_value = "New Value via PUT"
    shared_sme_string.value = new_value

    result = wrapper.put_submodel_element_by_path_submodel_repo(sm_id, shared_sme_string.id_short, shared_sme_string)

    assert result is True

    sme = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_string.id_short)

    assert sme is not None
    assert sme.id_short == shared_sme_string.id_short
    assert sme.value == new_value
    assert sme.value != old_value
    assert sme.description.get("en", "") == shared_sme_string.description.get("en", "")
    assert sme.display_name.get("en", "") == shared_sme_string.display_name.get("en", "")

    # restore original value
    shared_sme_string.value = "Sample String Value"
    wrapper.put_submodel_element_by_path_submodel_repo(sm_id, shared_sme_string.id_short, shared_sme_string)

def test_033a_get_submodel_element_by_path_value_only_submodel_repo(wrapper: SdkWrapper, shared_sme_string: model.Property):
    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    value = wrapper.get_submodel_element_by_path_value_only_submodel_repo(sm_id, shared_sme_string.id_short)

    parsed = urlparse(wrapper.base_url)
    if parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        assert value is None
        return

    assert value is not None

    submodel_element = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_string.id_short)
    assert submodel_element is not None
    prop: model.Property = submodel_element
    assert value == prop.value

def test_033b_get_submodel_element_by_path_value_only_submodel_repo(wrapper: SdkWrapper, shared_sme_int: model.Property):
    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    value = wrapper.get_submodel_element_by_path_value_only_submodel_repo(sm_id, shared_sme_int.id_short)

    parsed = urlparse(wrapper.base_url)
    if parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        assert value is None
        return

    assert value is not None

    submodel_element = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_int.id_short)
    assert submodel_element is not None
    prop: model.Property = submodel_element
    assert int(value) == prop.value

def test_033c_get_submodel_element_by_path_value_only_submodel_repo(wrapper: SdkWrapper, shared_sme_float: model.Property):
    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    value = wrapper.get_submodel_element_by_path_value_only_submodel_repo(sm_id, shared_sme_float.id_short)

    parsed = urlparse(wrapper.base_url)
    if parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        assert value is None
        return

    assert value is not None

    submodel_element = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_float.id_short)
    assert submodel_element is not None
    prop: model.Property = submodel_element
    assert float(value) == prop.value

def test_033d_get_submodel_element_by_path_value_only_submodel_repo(wrapper: SdkWrapper, shared_sme_bool: model.Property):
    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    value = wrapper.get_submodel_element_by_path_value_only_submodel_repo(sm_id, shared_sme_bool.id_short)

    parsed = urlparse(wrapper.base_url)
    if parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        assert value is None
        return

    assert value is not None

    sm_data = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_bool.id_short)
    assert sm_data is not None
    #assert bool(value) == bool(sm_data.get("value", ""))

def test_034_get_submodel_by_id_value_only(wrapper: SdkWrapper, shared_sm: model.Submodel):
    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    response = wrapper.get_submodel_by_id_value_only(sm_id)

    parsed = urlparse(wrapper.base_url)
    if parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        assert response is None
        return
    elif parsed.port in DOTNET_SERVER_PORTS:
        assert response is not None
        value = response[shared_sm.id_short]
    else:
        assert response is not None
        value = response

    assert value is not None
    assert len(value) > 3
    assert "sme_property_int" in value
    assert int(value["sme_property_int"]) == 263
    assert "sme_property_string" in value
    assert value["sme_property_string"] == "Sample String Value"
    assert "sme_property_float" in value
    assert float(value["sme_property_float"]) == 262.1

def test_035_patch_submodel_by_id_value_only(wrapper: SdkWrapper, shared_sm: model.Submodel, shared_sme_string: model.Property, shared_sme_int: model.Property, shared_sme_float: model.Property):
    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    value_dict = {
        shared_sme_string.id_short: shared_sme_string.value,
        shared_sme_int.id_short: str(shared_sme_int.value),
        shared_sme_float.id_short: str(shared_sme_float.value)
    }

    # patch_dict = {shared_sm.id: value_dict}

    patch_dict = value_dict

    parsed = urlparse(wrapper.base_url)
    if parsed.port in PYTHON_SERVER_PORTS:
        # NOTE: python server do not provide this endpoint
        return

    if parsed.port in JAVA_SERVER_PORTS:
        # NOTE: java server endpoint seems to work not correctly
        return

    elif parsed.port in DOTNET_SERVER_PORTS:
        patch_dict = value_dict

    result = wrapper.patch_submodel_by_id_value_only(sm_id, patch_dict)

    assert result is True

    string_prop_element = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_string.id_short)
    int_prop_element = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_int.id_short)
    float_prop_element = wrapper.get_submodel_element_by_path_submodel_repo(sm_id, shared_sme_float.id_short)

    assert string_prop_element is not None
    assert int_prop_element is not None
    assert float_prop_element is not None

    string_prop: model.Property = string_prop_element  # type: ignore
    int_prop: model.Property = int_prop_element  # type: ignore
    float_prop: model.Property = float_prop_element  # type: ignore

    assert string_prop.value == shared_sme_string.value
    assert int(int_prop.value) == int(shared_sme_int.value)
    assert float(float_prop.value) == float(shared_sme_float.value)

def test_036_get_submodel_by_id_metadata(wrapper: SdkWrapper, shared_sm: model.Submodel):
    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    metadata = wrapper.get_submodel_by_id_metadata(sm_id)
    assert metadata is not None

    submodel = wrapper.get_submodel_by_id(sm_id)
    assert submodel is not None

    assert metadata.get("id", "") == submodel.id
    assert metadata.get("idShort", "") == submodel.id_short
    assert metadata.get("description", {})[0].get("text", "") == submodel.description.get("en", "")
    assert metadata.get("displayName", {})[0].get("text", "") == submodel.display_name.get("en", "")
    assert "submodelElements" not in metadata

def test_098_delete_asset_administration_shell_by_id(wrapper: SdkWrapper):
    shell_id = SHELL_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        shell_id = encoder.encode_base_64(SHELL_ID)

    result = wrapper.delete_asset_administration_shell_by_id(shell_id)

    assert result

    shells = wrapper.get_all_asset_administration_shells()
    assert shells is not None
    assert len(shells.results) == 0

def test_099a_delete_submodel_by_id(wrapper: SdkWrapper):
    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(SM_ID)

    result = wrapper.delete_submodel_by_id(sm_id)

    assert result

    submodels = wrapper.get_all_submodels()
    assert submodels is not None
    assert len(submodels.results) == 1

def test_099b_delete_submodel_by_id(wrapper: SdkWrapper):
    sm_id = SM_ID

    if wrapper.get_encoded_ids() == IdEncoding.encoded:
        sm_id = encoder.encode_base_64(AIMC_SM_ID)

    result = wrapper.delete_submodel_by_id(sm_id)

    assert result

    submodels = wrapper.get_all_submodels()
    assert submodels is not None
    assert len(submodels.results) == 0



"""Utility functions for working with the BaSyx SDK framework objects."""

import json
import logging
from typing import Any

import basyx.aas.adapter.json
from basyx.aas import model

_logger = logging.getLogger(__name__)


def get_submodel_ids(shell: model.AssetAdministrationShell) -> list[str]:
    """Get all IDs from the submodels referenced in the given AAS.

    :param shell: The Asset Administration Shell to extract submodel IDs from.
    :return: A list of submodel IDs referenced in the AAS.
    """
    submodel_ids = []
    for submodel in shell.submodel:
        if len(submodel.key) < 1 or submodel.key[0].type != model.KeyTypes.SUBMODEL:
            _logger.warning(f"Submodel reference {submodel} does not start with SUBMODEL key type.")
            continue

        submodel_ids.append(submodel.key[0].value)

    return submodel_ids


def add_submodel_to_aas(aas: model.AssetAdministrationShell, submodel: model.Submodel) -> None:
    """Add a given Submodel correctly to a provided AssetAdministrationShell.

    :param aas: provided AssetAdministrationShell to which the Submodel should be added
    :param submodel: given Submodel to add
    """
    existing_submodel_ids = get_submodel_ids(aas)
    if submodel.id in existing_submodel_ids:
        _logger.warning(f"Submodel with ID {submodel.id} is already referenced in the AAS. Skipping addition.")
        return

    aas.submodel.add(model.ModelReference.from_referable(submodel))


def remove_submodel_from_aas(aas: model.AssetAdministrationShell, submodel: model.Submodel) -> None:
    """Remove a given Submodel correctly from a provided AssetAdministrationShell.

    :param aas: provided AssetAdministrationShell from which the Submodel should be removed
    :param submodel: given Submodel to remove
    """
    existing_submodel_ids = get_submodel_ids(aas)
    if submodel.id not in existing_submodel_ids:
        _logger.warning(f"Submodel with ID {submodel.id} is not referenced in the AAS. Skipping removal.")
        return

    aas.submodel.remove(model.ModelReference.from_referable(submodel))


def convert_to_object(content: dict) -> Any | None:
    """Convert a dictionary to a BaSyx SDK framework object.

    :param content: dictionary to convert
    :return: BaSyx SDK framework object or None
    """
    if not content or len(content) == 0:
        _logger.debug("Empty content provided for conversion to object.")
        return None

    try:
        dict_string = json.dumps(content)
        return json.loads(dict_string, cls=basyx.aas.adapter.json.json_deserialization.AASFromJsonDecoder)
    except Exception as e:
        _logger.error(f"Decoding error: {e}")
        _logger.error(f"In JSON: {content}")
        return None


def convert_to_dict(object: Any) -> dict | None:
    """Convert a BaSyx SDK framework object. to a dictionary.

    :param object: BaSyx SDK framework object to convert
    :return: dictionary representation of the object or None
    """
    if not object:
        _logger.debug("Empty object provided for conversion to dictionary.")
        return None

    try:
        data_string = json.dumps(object, cls=basyx.aas.adapter.json.AASToJsonEncoder)
        model_dict = json.loads(data_string)
        return model_dict
    except Exception as e:
        _logger.error(f"Encoding error: {e}")
        _logger.error(f"In object: {object}")
        return None



"""Client Implementations Module."""

from aas_http_client.classes.client.implementations.authentication import AuthMethod, TokenData, get_token
from aas_http_client.classes.client.implementations.experimental_implementation import ExperimentalImplementation
from aas_http_client.classes.client.implementations.shell_implementation import ShellRepoImplementation
from aas_http_client.classes.client.implementations.shell_registry_implementation import ShellRegistryImplementation
from aas_http_client.classes.client.implementations.sm_implementation import SubmodelRepoImplementation
from aas_http_client.classes.client.implementations.sm_registry_implementation import SubmodelRegistryImplementation

__all__ = [
    "AuthMethod",
    "ExperimentalImplementation",
    "ShellRegistryImplementation",
    "ShellRepoImplementation",
    "SubmodelRegistryImplementation",
    "SubmodelRepoImplementation",
    "TokenData",
    "get_token",
]



"""Implementation of Asset Administration Shell related API calls."""

import json
import logging
import mimetypes
from pathlib import Path
from typing import TYPE_CHECKING, Any

import requests
from pydantic import BaseModel

if TYPE_CHECKING:
    from aas_http_client.classes.client.aas_client import AasHttpClient

from aas_http_client.utilities.encoder import encode_base_64
from aas_http_client.utilities.http_helper import (
    STATUS_CODE_200,
    STATUS_CODE_201,
    STATUS_CODE_204,
    STATUS_CODE_404,
    log_response,
)

_logger = logging.getLogger(__name__)


class ShellRepoImplementation(BaseModel):
    """Implementation of Asset Administration Shell related API calls."""

    def __init__(self, client: "AasHttpClient"):
        """Initializes the ShellImplementation with the given parameters."""
        self._client = client

        session = client.get_session()
        if session is None:
            raise ValueError(
                "HTTP session is not initialized in the client. Call 'initialize()' method of the client before creating SubmodelRegistryImplementation instance."
            )

        self._session: requests.Session = session

    # GET /shells/{aasIdentifier}
    def get_asset_administration_shell_by_id(self, aas_identifier: str) -> dict | None:
        """Returns a specific Asset Administration Shell.

        :param aas_identifier: The Asset Administration Shells unique id
        :return: Asset Administration Shells data or None if an error occurred
        """
        if not self._client.encoded_ids:
            aas_identifier = encode_base_64(aas_identifier)

        url = f"{self._client.base_url}/shells/{aas_identifier}"

        self._client.set_token()

        try:
            response = self._session.get(url, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Asset Administration Shell with id '{aas_identifier}' not found.")
                _logger.debug(response.text)
                return None

            if response.status_code != STATUS_CODE_200:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        content = response.content.decode("utf-8")
        return json.loads(content)

    # PUT /shells/{aasIdentifier}
    def put_asset_administration_shell_by_id(self, aas_identifier: str, request_body: dict) -> bool:
        """Creates or replaces an existing Asset Administration Shell.

        :param aas_identifier: The Asset Administration Shells unique id
        :param request_body: Json data of the Asset Administration Shell data to put
        :return: True if the update was successful, False otherwise
        """
        if not self._client.encoded_ids:
            aas_identifier = encode_base_64(aas_identifier)

        url = f"{self._client.base_url}/shells/{aas_identifier}"

        self._client.set_token()

        try:
            response = self._session.put(url, json=request_body, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Asset Administration Shell with id '{aas_identifier}' not found.")
                _logger.debug(response.text)
                return False

            if response.status_code is not STATUS_CODE_204:
                log_response(response)
                return False

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return False

        return True

    # DELETE /shells/{aasIdentifier}
    def delete_asset_administration_shell_by_id(self, aas_identifier: str) -> bool:
        """Deletes an Asset Administration Shell.

        :param aas_identifier: The Asset Administration Shells unique id
        :return: True if the deletion was successful, False otherwise
        """
        if not self._client.encoded_ids:
            aas_identifier = encode_base_64(aas_identifier)

        url = f"{self._client.base_url}/shells/{aas_identifier}"

        self._client.set_token()

        try:
            response = self._session.delete(url, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Asset Administration Shell with id '{aas_identifier}' not found.")
                _logger.debug(response.text)
                return False

            if response.status_code != STATUS_CODE_204:
                log_response(response)
                return False

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return False

        return True

    # GET /shells/{aasIdentifier}/asset-information/thumbnail
    def get_thumbnail_aas_repository(self, aas_identifier: str) -> bytes | None:
        """Returns the thumbnail of the Asset Administration Shell.

        :param aas_identifier: The Asset Administration Shells unique id
        :return: Thumbnail file data as bytes (octet-stream) or None if an error occurred
        """
        if not self._client.encoded_ids:
            aas_identifier = encode_base_64(aas_identifier)

        url = f"{self._client.base_url}/shells/{aas_identifier}/asset-information/thumbnail"

        self._client.set_token()

        try:
            response = self._session.get(url, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Asset Administration Shell with id '{aas_identifier}' or thumbnail file not found.")
                _logger.debug(response.text)
                return None

            if response.status_code != STATUS_CODE_200:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        return response.content

    # PUT /shells/{aasIdentifier}/asset-information/thumbnail
    def put_thumbnail_aas_repository(self, aas_identifier: str, file_name: str, file: Path) -> bool:
        """Creates or updates the thumbnail of the Asset Administration Shell.

        :param aas_identifier: The Asset Administration Shells unique id
        :param file_name: The name of the thumbnail file
        :param file: Path to the thumbnail file to upload as attachment
        :return: True if the update was successful, False otherwise
        """
        if file.exists() is False or not file.is_file():
            _logger.error(f"Attachment file '{file}' does not exist.")
            return False

        if not self._client.encoded_ids:
            aas_identifier = encode_base_64(aas_identifier)

        url = f"{self._client.base_url}/shells/{aas_identifier}/asset-information/thumbnail"

        params = {"fileName": file_name}

        self._client.set_token()

        try:
            mime_type, _ = mimetypes.guess_type(file)

            with file.open("rb") as f:
                files = {"file": (file.name, f, mime_type or "application/octet-stream")}
                response = self._session.put(url, files=files, params=params, timeout=self._client.time_out)

            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Asset Administration Shell with id '{aas_identifier}' not found.")
                _logger.debug(response.text)
                return False

            # original dotnet server delivers 200 instead of 204
            if response.status_code not in (STATUS_CODE_200, STATUS_CODE_204):
                log_response(response)
                return False

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return False

        return True

    # DELETE /shells/{aasIdentifier}/asset-information/thumbnail
    def delete_thumbnail_aas_repository(self, aas_identifier: str) -> bool:
        """Deletes the thumbnail of the Asset Administration Shell.

        :param aas_identifier: The Asset Administration Shells unique id
        :return: True if the deletion was successful, False otherwise
        """
        if not self._client.encoded_ids:
            aas_identifier = encode_base_64(aas_identifier)

        url = f"{self._client.base_url}/shells/{aas_identifier}/asset-information/thumbnail"

        self._client.set_token()

        try:
            response = self._session.delete(url, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Asset Administration Shell with id '{aas_identifier}' or thumbnail file not found.")
                _logger.debug(response.text)
                return False

            if response.status_code != STATUS_CODE_200:
                log_response(response)
                return False

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return False

        return True

    # GET /shells
    def get_all_asset_administration_shells(
        self, asset_ids: list[dict] | None = None, id_short: str = "", limit: int = 100, cursor: str = ""
    ) -> dict | None:
        """Returns all Asset Administration Shells.

        :param assetIds: A list of specific Asset identifiers (format: {"identifier": "string",  "encodedIdentifier": "string"})
        :param idShort: The Asset Administration Shells IdShort
        :param limit: The maximum number of elements in the response array
        :param cursor: A server-generated identifier retrieved from pagingMetadata that specifies from which position the result listing should continue
        :return: List of paginated Asset Administration Shells data or None if an error occurred
        """
        url = f"{self._client.base_url}/shells"

        # Build query parameters
        if asset_ids is None:
            asset_ids = []

        params: dict[str, Any] = {}
        if asset_ids is not None and len(asset_ids) > 0:
            params["assetIds"] = asset_ids
        if id_short:
            params["idShort"] = id_short
        if limit:
            params["limit"] = str(limit)
        if cursor:
            params["cursor"] = cursor

        self._client.set_token()

        try:
            response = self._session.get(url, timeout=self._client.time_out, params=params)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code != STATUS_CODE_200:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        content = response.content.decode("utf-8")
        return json.loads(content)

    # POST /shells
    def post_asset_administration_shell(self, request_body: dict) -> dict | None:
        """Creates a new Asset Administration Shell.

        :param request_body: Json data of the Asset Administration Shell to post
        :return: Response data as a dictionary or None if an error occurred
        """
        url = f"{self._client.base_url}/shells"

        self._client.set_token()

        try:
            response = self._session.post(url, json=request_body, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code != STATUS_CODE_201:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        content = response.content.decode("utf-8")
        return json.loads(content)

    # GET /shells/{aasIdentifier}/submodel-refs
    def get_all_submodel_references_aas_repository(self, aas_identifier: str, limit: int = 100, cursor: str = "") -> dict | None:
        """Returns all submodel references.

        :param aas_identifier: The Asset Administration Shells unique id
        :param limit: The maximum number of elements in the response array
        :param cursor: A server-generated identifier retrieved from pagingMetadata that specifies from which position the result listing should continue
        :return: List of Submodel references or None if an error occurred
        """
        if not self._client.encoded_ids:
            aas_identifier = encode_base_64(aas_identifier)

        url = f"{self._client.base_url}/shells/{aas_identifier}/submodel-refs"

        params: dict[str, str] = {}
        if limit:
            params["limit"] = str(limit)
        if cursor:
            params["cursor"] = cursor

        self._client.set_token()

        try:
            response = self._session.get(url, timeout=self._client.time_out, params=params)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Asset Administration Shell with id '{aas_identifier}' not found.")
                _logger.debug(response.text)
                return None

            if response.status_code != STATUS_CODE_200:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        content = response.content.decode("utf-8")
        return json.loads(content)

    # POST /shells/{aasIdentifier}/submodel-refs
    def post_submodel_reference_aas_repository(self, aas_identifier: str, request_body: dict) -> dict | None:
        """Creates a submodel reference at the Asset Administration Shell.

        :param aas_identifier: The Asset Administration Shells unique id
        :param request_body: Reference to the Submodel
        :return: Response data as a dictionary or None if an error occurred
        """
        if not self._client.encoded_ids:
            aas_identifier = encode_base_64(aas_identifier)

        url = f"{self._client.base_url}/shells/{aas_identifier}/submodel-refs"

        self._client.set_token()

        try:
            response = self._session.post(url, json=request_body, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Asset Administration Shell with id '{aas_identifier}' not found.")
                _logger.debug(response.text)
                return None

            if response.status_code != STATUS_CODE_201:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        content = response.content.decode("utf-8")
        return json.loads(content)

    # DELETE /shells/{aasIdentifier}/submodel-refs/{submodelIdentifier}
    def delete_submodel_reference_by_id_aas_repository(self, aas_identifier: str, submodel_identifier: str) -> bool:
        """Deletes the submodel reference from the Asset Administration Shell. Does not delete the submodel itself.

        :param aas_identifier: The Asset Administration Shells unique id
        :param submodel_identifier: The Submodels unique id
        :return: True if the deletion was successful, False otherwise
        """
        if not self._client.encoded_ids:
            aas_identifier = encode_base_64(aas_identifier)
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/shells/{aas_identifier}/submodel-refs/{submodel_identifier}"

        self._client.set_token()

        try:
            response = self._session.delete(url, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Asset Administration Shell with id '{aas_identifier}' or submodel with id '{submodel_identifier}' not found.")
                _logger.debug(response.text)
                return False

            if response.status_code not in (STATUS_CODE_204, STATUS_CODE_200):
                log_response(response)
                return False

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return False

        return True

    # not supported by Java Server

    # PUT /shells/{aasIdentifier}/submodels/{submodelIdentifier}
    def put_submodel_by_id_aas_repository(self, aas_identifier: str, submodel_identifier: str, request_body: dict) -> bool:
        """Updates the Submodel.

        :param aas_identifier: ID of the AAS to update the submodel for
        :param submodel_identifier: ID of the submodel to update
        :param request_body: Json data to the Submodel to put
        :return: True if the update was successful, False otherwise
        """
        if not self._client.encoded_ids:
            aas_identifier = encode_base_64(aas_identifier)
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/shells/{aas_identifier}/submodels/{submodel_identifier}"

        self._client.set_token()

        try:
            response = self._session.put(url, json=request_body, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Asset Administration Shell with id '{aas_identifier}' or submodel with id '{submodel_identifier}' not found.")
                _logger.debug(response.text)
                return False

            if response.status_code != STATUS_CODE_204:
                log_response(response)
                return False

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return False

        return True

    # GET /shells/{aasIdentifier}/$reference
    def get_asset_administration_shell_by_id_reference_aas_repository(self, aas_identifier: str) -> dict | None:
        """Returns a specific Asset Administration Shell as a Reference.

        :param aas_identifier: ID of the AAS reference to retrieve
        :return: Asset Administration Shells reference data or None if an error occurred
        """
        if not self._client.encoded_ids:
            aas_identifier = encode_base_64(aas_identifier)

        url = f"{self._client.base_url}/shells/{aas_identifier}/$reference"

        self._client.set_token()

        try:
            response = self._session.get(url, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Asset Administration Shell with id '{aas_identifier}' not found.")
                _logger.debug(response.text)
                return None

            if response.status_code != STATUS_CODE_200:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        ref_dict_string = response.content.decode("utf-8")
        return json.loads(ref_dict_string)

    # GET /shells/{aasIdentifier}/submodels/{submodelIdentifier}
    def get_submodel_by_id_aas_repository(self, aas_identifier: str, submodel_identifier: str) -> dict | None:
        """Returns the Submodel.

        :param aas_identifier: ID of the AAS to retrieve the submodel from
        :param submodel_identifier: ID of the submodel to retrieve
        :return: Submodel object or None if an error occurred
        """
        if not self._client.encoded_ids:
            aas_identifier = encode_base_64(aas_identifier)
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/shells/{aas_identifier}/submodels/{submodel_identifier}"

        self._client.set_token()

        try:
            response = self._session.get(url, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Asset Administration Shell with id '{aas_identifier}' or submodel with id '{submodel_identifier}' not found.")
                _logger.debug(response.text)
                return None

            if response.status_code != STATUS_CODE_200:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return None

        content = response.content.decode("utf-8")
        return json.loads(content)



"""Main process for the demo."""

import json
import logging
from pathlib import Path

from basyx.aas import model

from aas_http_client.classes.client import aas_client
from aas_http_client.classes.wrapper import sdk_wrapper
from aas_http_client.utilities import encoder, model_builder, sdk_tools

logger = logging.getLogger(__name__)


def start() -> None:
    """Start the demo process."""

    # create a submodel element
    wrapper = sdk_wrapper.create_by_url(base_url="http://javaaasserver:8075", encoded_ids=False)
    client = wrapper.get_client()
    # client = aas_client.create_client_by_url(base_url="http://pythonaasserver:80/", encoded_ids=False)
    client_shell_reg = aas_client.create_by_url(base_url="http://aas-registry:8080", encoded_ids=False)

    sm = model_builder.create_base_submodel("TestSubmodel", "TestSM")
    shell = model_builder.create_base_aas("TestAAS", "TestAAS")
    sdk_tools.add_submodel_to_aas(shell, sm)

    iqstrcut = aas_client.create_by_url(
        base_url="https://aurora-fluid40-rt.iqstruct-engineering.de/aas-env/",
        o_auth_client_id="fluid40",
        o_auth_client_secret="LdFB4jRrMMkgcVWgFkOVdDVDXtQ5os8w",
        o_auth_token_url="https://aurora-fluid40-rt.iqstruct-engineering.de/auth/realms/BaSyx/protocol/openid-connect/token",
        encoded_ids=False,
    )

    client.shells.delete_asset_administration_shell_by_id(shell.id)
    client.submodels.delete_submodel_by_id(sm.id)

    shell_data = sdk_tools.convert_to_dict(shell)
    sm_data = sdk_tools.convert_to_dict(sm)

    result = client.submodels.post_submodel(sm_data)
    result = client.shells.post_asset_administration_shell(shell_data)

    put_result = client.shells.put_thumbnail_aas_repository(shell.id, "Pen_Machine.png", Path("./tests/test_data/Pen_Machine.png").resolve())

    descriptors = client_shell_reg.shell_registry.get_all_asset_administration_shell_descriptors()
    descriptor = sdk_tools.convert_to_object(descriptors.get("result", [])[0])

    file_sme = model.File("file_sme", content_type="application/pdf")
    file_post_result = client.submodels.post_submodel_element_submodel_repo(sm.id, sdk_tools.convert_to_dict(file_sme))

    file_get_result = wrapper.get_submodel_element_by_path_submodel_repo(sm.id, file_sme.id_short)

    file = Path(f"./tests/test_data/https.pdf").resolve()
    tmp = client.experimental.post_file_by_path_submodel_repo(sm.id, file_sme.id_short, file)

    file_get_result = client.submodels.get_submodel_element_by_path_submodel_repo(sm.id, file_sme.id_short)

    attachment = client.experimental.get_file_by_path_submodel_repo(sm.id, file_sme.id_short)

    attachment_2 = wrapper.experimental_get_file_by_path_submodel_repo(sm.id, file_sme.id_short)

    file = Path(f"./tests/test_data/aimc.json").resolve()
    tmp = client.experimental.put_file_by_path_submodel_repo(sm.id, file_sme.id_short, file)

    attachment = client.experimental.get_file_by_path_submodel_repo(sm.id, file_sme.id_short)

    delete_result = client.experimental.delete_file_by_path_submodel_repo(sm.id, file_sme.id_short)

    attachment = client.experimental.get_file_by_path_submodel_repo(sm.id, file_sme.id_short)

    print(attachment)



"""Client for HTTP API communication with AAS server."""

import json
import logging
import time
from pathlib import Path
from typing import Any

import requests
from pydantic import BaseModel, Field, PrivateAttr, ValidationError
from requests import Session
from requests.auth import HTTPBasicAuth

from aas_http_client.classes.client.implementations import (
    AuthMethod,
    ExperimentalImplementation,
    ShellRegistryImplementation,
    ShellRepoImplementation,
    SubmodelRegistryImplementation,
    SubmodelRepoImplementation,
    TokenData,
    get_token,
)
from aas_http_client.classes.Configuration.config_classes import AuthenticationConfig
from aas_http_client.utilities.http_helper import (
    STATUS_CODE_200,
    STATUS_CODE_201,
    STATUS_CODE_202,
    STATUS_CODE_204,
)

_logger = logging.getLogger(__name__)


class AasHttpClient(BaseModel):
    """Represents a AasHttpClient to communicate with a REST API."""

    base_url: str = Field(..., alias="BaseUrl", description="Base URL of the AAS server.")
    auth_settings: AuthenticationConfig = Field(
        default_factory=AuthenticationConfig, alias="AuthenticationSettings", description="Authentication settings for the AAS server."
    )
    https_proxy: str | None = Field(default=None, alias="HttpsProxy", description="HTTPS proxy URL.")
    http_proxy: str | None = Field(default=None, alias="HttpProxy", description="HTTP proxy URL.")
    time_out: int = Field(default=200, alias="TimeOut", description="Timeout for HTTP requests.")
    connection_time_out: int = Field(default=100, alias="ConnectionTimeOut", description="Connection timeout for HTTP requests.")
    ssl_verify: bool = Field(default=True, alias="SslVerify", description="Enable SSL verification.")
    trust_env: bool = Field(default=True, alias="TrustEnv", description="Trust environment variables.")
    _session: Session | None = PrivateAttr(default=None)
    _auth_method: AuthMethod = PrivateAttr(default=AuthMethod.basic_auth)
    encoded_ids: bool = Field(default=True, alias="EncodedIds", description="If enabled, all IDs used in API requests have to be base64-encoded.")
    shells: ShellRepoImplementation | None = Field(default=None)
    submodels: SubmodelRepoImplementation | None = Field(default=None)
    shell_registry: ShellRegistryImplementation | None = Field(default=None)
    experimental: ExperimentalImplementation | None = Field(default=None)
    submodel_registry: SubmodelRegistryImplementation | None = Field(default=None)
    _cached_token: TokenData | None = PrivateAttr(default=None)

    def initialize(self):
        """Initialize the AasHttpClient with the given URL, username and password."""
        if self.base_url.endswith("/"):
            self.base_url = self.base_url[:-1]

        self._session = requests.Session()

        self._handle_auth_method()

        self._session.verify = self.ssl_verify
        self._session.trust_env = self.trust_env

        if self.https_proxy:
            self._session.proxies.update({"https": self.https_proxy})
        if self.http_proxy:
            self._session.proxies.update({"http": self.http_proxy})

        self._session.headers.update(
            {
                "Accept": "*/*",
                "User-Agent": "python-requests/2.32.5",
                "Connection": "close",
            }
        )

        self.shells = ShellRepoImplementation(self)
        self.submodels = SubmodelRepoImplementation(self)
        self.shell_registry = ShellRegistryImplementation(self)
        self.submodel_registry = SubmodelRegistryImplementation(self)
        self.experimental = ExperimentalImplementation(self)

    def get_session(self) -> Session | None:
        """Get the HTTP session used by the client.

        :return: The requests.Session object used for HTTP communication
        """
        return self._session

    def _handle_auth_method(self):
        """Handles the authentication method based on the provided settings."""
        if self.auth_settings.o_auth.is_active():
            self._auth_method = AuthMethod.o_auth
            _logger.info(
                f"Authentication method: OAuth | '{self.auth_settings.o_auth.client_id}' | '{self.auth_settings.o_auth.token_url}' | '{self.auth_settings.o_auth.grant_type}'"
            )

        elif self.auth_settings.basic_auth.is_active():
            self._auth_method = AuthMethod.basic_auth
            _logger.info(f"Authentication method: Basic Auth | '{self.auth_settings.basic_auth.username}'")
            self._session.auth = HTTPBasicAuth(self.auth_settings.basic_auth.username, self.auth_settings.basic_auth.get_password())

        elif self.auth_settings.bearer_auth.is_active():
            self._auth_method = AuthMethod.bearer
            _logger.info("Authentication method: Bearer Token")
            self._session.headers.update({"Authorization": f"Bearer {self.auth_settings.bearer_auth.get_token()}"})

        else:
            self._auth_method = AuthMethod.No
            _logger.info("Authentication method: No Authentication")

    def get_root(self) -> dict | None:
        """Get the root endpoint of the AAS server API to test connectivity.

        This method calls the '/shells' endpoint to verify that the AAS server is accessible
        and responding. It automatically handles authentication token setup if service
        provider authentication is configured.

        :return: Response data as a dictionary containing shell information, or None if an error occurred
        """
        if not self._session:
            _logger.error("HTTP session is not initialized. Call 'initialize()' method before making API calls.")
            return None

        urls: list[str] = []
        urls.append(f"{self.base_url}/shells")
        urls.append(f"{self.base_url}/submodels")
        urls.append(f"{self.base_url}/shell-descriptors")
        urls.append(f"{self.base_url}/submodel-descriptors")

        self.set_token()

        for url in urls:
            _logger.debug(f"Testing connectivity with URL: {url}")
            try:
                response = self._session.get(url, timeout=10)
                _logger.debug(f"Call REST API url '{response.url}'")

                if response.status_code == STATUS_CODE_200:
                    content = response.content.decode("utf-8")
                    return json.loads(content)

            except requests.exceptions.RequestException as e:
                _logger.error(f"Error call REST API: {e}")

        return None

    def set_token(self) -> str | None:
        """Set authentication token in session headers based on configured authentication method.

        :return: The access token if set, otherwise None
        """
        if not self._session:
            _logger.error("HTTP session is not initialized. Call 'initialize()' method before making API calls.")
            return None

        if self._auth_method != AuthMethod.o_auth:
            return None

        now = time.time()
        # Check if cached token exists and is not expired
        if self._cached_token and self._cached_token.token_expiry > now:
            return self._cached_token.access_token

        # Obtain new token
        token_data = get_token(self.auth_settings.o_auth)

        if token_data and token_data.access_token:
            # Cache the token data
            self._cached_token = token_data
            # Update session headers with the new token
            self._session.headers.update({"Authorization": f"Bearer {self._cached_token.access_token}"})
            return self._cached_token.access_token

        return None

    def get_endpoint(self, end_point_url: str) -> None | dict:
        """Generic GET request for endpoint.

        :param end_point_url: The endpoint URL to send the GET request to.
        :return: The base URL of the AAS server.
        """
        if not self._session:
            _logger.error("HTTP session is not initialized. Call 'initialize()' method before making API calls.")
            return None

        try:
            response = self._session.get(end_point_url, timeout=self.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_200:
                content = response.content.decode("utf-8")
                return json.loads(content)

        except requests.exceptions.RequestException as e:
            _logger.debug(f"Error call REST API: {e}")

        return None

    def put_endpoint(self, end_point_url: str, request_body: dict) -> None | dict:
        """Generic PUT request for endpoint.

        :param end_point_url: The endpoint URL to send the PUT request to.
        :param request_body: The request body to send with the PUT request.
        :return: The base URL of the AAS server.
        """
        if not self._session:
            _logger.error("HTTP session is not initialized. Call 'initialize()' method before making API calls.")
            return None

        try:
            response = self._session.put(end_point_url, json=request_body, timeout=self.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code not in (STATUS_CODE_200, STATUS_CODE_201, STATUS_CODE_204):
                content = response.content.decode("utf-8")
                return json.loads(content)

        except requests.exceptions.RequestException as e:
            _logger.debug(f"Error call REST API: {e}")

        return None

    def post_endpoint(self, end_point_url: str, request_body: dict) -> None | dict:
        """Generic POST request for endpoint.

        :param end_point_url: The endpoint URL to send the POST request to.
        :param request_body: The request body to send with the POST request.
        :return: The base URL of the AAS server.
        """
        if not self._session:
            _logger.error("HTTP session is not initialized. Call 'initialize()' method before making API calls.")
            return None

        try:
            response = self._session.post(end_point_url, json=request_body, timeout=self.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code not in (STATUS_CODE_201, STATUS_CODE_200, STATUS_CODE_202):
                content = response.content.decode("utf-8")
                return json.loads(content)

        except requests.exceptions.RequestException as e:
            _logger.debug(f"Error call REST API: {e}")

        return None

    def patch_endpoint(self, end_point_url: str, request_body: dict) -> None | dict:
        """Generic PATCH request for endpoint.

        :param end_point_url: The endpoint URL to send the PATCH request to.
        :param request_body: The request body to send with the PATCH request.
        :return: The base URL of the AAS server.
        """
        if not self._session:
            _logger.error("HTTP session is not initialized. Call 'initialize()' method before making API calls.")
            return None

        try:
            response = self._session.patch(end_point_url, json=request_body, timeout=self.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code not in (STATUS_CODE_200, STATUS_CODE_204):
                content = response.content.decode("utf-8")
                return json.loads(content)

        except requests.exceptions.RequestException as e:
            _logger.debug(f"Error call REST API: {e}")

        return None

    def delete_endpoint(self, end_point_url: str) -> None | dict:
        """Generic DELETE request for endpoint.

        :param end_point_url: The endpoint URL to send the DELETE request to.
        :return: The base URL of the AAS server.
        """
        if not self._session:
            _logger.error("HTTP session is not initialized. Call 'initialize()' method before making API calls.")
            return None

        try:
            response = self._session.delete(end_point_url, timeout=self.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code not in (STATUS_CODE_200, STATUS_CODE_204, STATUS_CODE_202):
                content = response.content.decode("utf-8")
                return json.loads(content)

        except requests.exceptions.RequestException as e:
            _logger.debug(f"Error call REST API: {e}")

        return None


def create_by_url(  # noqa: PLR0913
    base_url: str,
    basic_auth_username: str = "",
    basic_auth_password: str = "",
    o_auth_client_id: str = "",
    o_auth_client_secret: str = "",
    o_auth_token_url: str = "",
    bearer_auth_token: str = "",
    http_proxy: str = "",
    https_proxy: str = "",
    time_out: int = 200,
    connection_time_out: int = 60,
    ssl_verify: bool = True,  # noqa: FBT001, FBT002
    trust_env: bool = True,  # noqa: FBT001, FBT002
    encoded_ids: bool = True,  # noqa: FBT001, FBT002
) -> AasHttpClient | None:
    """Create a HTTP client for a AAS server connection from the given parameters.

    :param base_url: Base URL of the AAS server, e.g. "http://basyx_python_server:80/"
    :param basic_auth_username: Username for the AAS server basic authentication, defaults to ""
    :param basic_auth_password: Password for the AAS server basic authentication, defaults to ""
    :param o_auth_client_id: Client ID for OAuth authentication, defaults to ""
    :param o_auth_client_secret: Client secret for OAuth authentication, defaults to ""
    :param o_auth_token_url: Token URL for OAuth authentication, defaults to ""
    :param bearer_auth_token: Bearer token for authentication, defaults to ""
    :param http_proxy: HTTP proxy URL, defaults to ""
    :param https_proxy: HTTPS proxy URL, defaults to ""
    :param time_out: Timeout for the API calls, defaults to 200
    :param connection_time_out: Timeout for the connection to the API, defaults to 60
    :param ssl_verify: Whether to verify SSL certificates, defaults to True
    :param trust_env: Whether to trust environment variables for proxy settings, defaults to True
    :param encoded_ids: If enabled, all IDs used in API requests have to be base64-encoded
    :return: An instance of AasHttpClient initialized with the provided parameters or None if connection fails
    """
    _logger.info(f"Create AAS server http client from URL '{base_url}'.")
    config_dict: dict[str, Any] = {}
    config_dict["BaseUrl"] = base_url
    config_dict["HttpProxy"] = http_proxy
    config_dict["HttpsProxy"] = https_proxy
    config_dict["TimeOut"] = str(time_out)
    config_dict["ConnectionTimeOut"] = str(connection_time_out)
    config_dict["SslVerify"] = str(ssl_verify)
    config_dict["TrustEnv"] = str(trust_env)
    config_dict["EncodedIds"] = str(encoded_ids)

    config_dict["AuthenticationSettings"] = {
        "BasicAuth": {"Username": basic_auth_username},
        "OAuth": {
            "ClientId": o_auth_client_id,
            "TokenUrl": o_auth_token_url,
        },
    }

    return create_by_dict(config_dict, basic_auth_password, o_auth_client_secret, bearer_auth_token)


def create_by_dict(
    configuration: dict, basic_auth_password: str = "", o_auth_client_secret: str = "", bearer_auth_token: str = ""
) -> AasHttpClient | None:
    """Create a HTTP client for a AAS server connection from the given configuration.

    :param configuration: Dictionary containing the AAS server connection settings
    :param basic_auth_password: Password for the AAS server basic authentication, defaults to ""
    :param o_auth_client_secret: Client secret for OAuth authentication, defaults to ""
    :param bearer_auth_token: Bearer token for authentication, defaults to ""
    :return: An instance of AasHttpClient initialized with the provided parameters or None if validation fails
    """
    _logger.info("Create AAS server http client from dictionary.")
    config_string = json.dumps(configuration, indent=4)

    return _create_client(config_string, basic_auth_password, o_auth_client_secret, bearer_auth_token)


def create_by_config(
    config_file: Path, basic_auth_password: str = "", o_auth_client_secret: str = "", bearer_auth_token: str = ""
) -> AasHttpClient | None:
    """Create a HTTP client for a AAS server connection from a given configuration file.

    :param config_file: Path to the configuration file containing the AAS server connection settings
    :param basic_auth_password: Password for the AAS server basic authentication, defaults to ""
    :param o_auth_client_secret: Client secret for OAuth authentication, defaults to ""
    :param bearer_auth_token: Bearer token for authentication, defaults to ""
    :return: An instance of AasHttpClient initialized with the provided parameters or None if validation fails
    """
    config_file = config_file.resolve()
    _logger.info(f"Create AAS server http client from configuration file '{config_file}'.")
    if not config_file.exists():
        config_string = "{}"
        _logger.warning(f"Configuration file '{config_file}' not found. Using default configuration.")
    else:
        config_string = config_file.read_text(encoding="utf-8")
        _logger.debug(f"Configuration  file '{config_file}' found.")

    return _create_client(config_string, basic_auth_password, o_auth_client_secret, bearer_auth_token)


def _create_client(config_string: str, basic_auth_password: str, o_auth_client_secret: str, bearer_auth_token: str) -> AasHttpClient | None:
    """Create and initialize an AAS HTTP client from configuration string.

    This internal method validates the configuration, sets authentication credentials,
    initializes the client, and tests the connection to the AAS server.

    :param config_string: JSON configuration string containing AAS server settings
    :param basic_auth_password: Password for basic authentication, defaults to ""
    :param o_auth_client_secret: Client secret for OAuth authentication, defaults to ""
    :param bearer_auth_token: Bearer token for authentication, defaults to ""
    :return: An initialized and connected AasHttpClient instance or None if connection fails
    :raises ValidationError: If the configuration string is invalid
    :raises TimeoutError: If connection to the server times out
    """
    try:
        client = AasHttpClient.model_validate_json(config_string)
    except ValidationError as ve:
        raise ValidationError(f"Invalid BaSyx server configuration file: {ve}") from ve

    client.auth_settings.basic_auth.set_password(basic_auth_password)
    client.auth_settings.o_auth.set_client_secret(o_auth_client_secret)
    client.auth_settings.bearer_auth.set_token(bearer_auth_token)

    _logger.info("Using server configuration:")
    _logger.info(f"BaseUrl: '{client.base_url}'")
    _logger.info(f"TimeOut: '{client.time_out}'")
    _logger.info(f"HttpsProxy: '{client.https_proxy}'")
    _logger.info(f"HttpProxy: '{client.http_proxy}'")
    _logger.info(f"ConnectionTimeOut: '{client.connection_time_out}'.")
    _logger.info(f"SSLVerify: '{client.ssl_verify}'.")
    _logger.info(f"TrustEnv: '{client.trust_env}'.")
    _logger.info(f"EncodedIds: '{client.encoded_ids}'.")

    client.initialize()

    # test the connection to the REST API
    connected = __connect_to_api(client)

    if not connected:
        return None

    return client


def __connect_to_api(client: AasHttpClient) -> bool:
    """Test the connection to the AAS server API with retry logic.

    This internal method attempts to establish a connection to the AAS server by calling
    the get_root() method. It retries the connection for the duration specified in the
    client's connection_time_out setting, sleeping 1 second between attempts.

    :param client: The AasHttpClient instance to test the connection for
    :return: True if connection is successful, False otherwise
    :raises TimeoutError: If connection attempts fail for the entire timeout duration
    """
    start_time = time.time()
    _logger.info(f"Try to connect to REST API '{client.base_url}' for {client.connection_time_out} seconds.")
    counter: int = 0
    while True:
        try:
            root = client.get_root()
            if root:
                _logger.info(f"Connected to server API at '{client.base_url}' successfully.")
                return True

            _logger.error(f"Connection attempt to '{client.base_url}' failed.")

        except requests.exceptions.ConnectionError:
            pass
        if time.time() - start_time > client.connection_time_out:
            raise TimeoutError(f"Connection to server API timed out after {client.connection_time_out} seconds.")

        counter += 1
        _logger.warning(f"Retrying connection (attempt: {counter}).")
        time.sleep(1)



"""Experimental implementation of Asset Administration Shell Registry related API calls."""

import logging
import mimetypes
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aas_http_client.classes.client.aas_client import AasHttpClient

import requests
from pydantic import BaseModel

from aas_http_client.utilities.encoder import encode_base_64
from aas_http_client.utilities.http_helper import (
    STATUS_CODE_200,
    STATUS_CODE_204,
    STATUS_CODE_404,
    log_response,
)

_logger = logging.getLogger(__name__)


class ExperimentalImplementation(BaseModel):
    """Implementation of Asset Administration Shell Registry related API calls."""

    def __init__(self, client: "AasHttpClient"):
        """Initializes the ExperimentalImplementation with the given client."""
        self._client = client

        session = client.get_session()
        if session is None:
            raise ValueError(
                "HTTP session is not initialized in the client. Call 'initialize()' method of the client before creating SubmodelRegistryImplementation instance."
            )

        self._session: requests.Session = session

    # GET /submodels/{submodelIdentifier}/submodel-elements/{idShortPath}/attachment
    def get_file_by_path_submodel_repo(self, submodel_identifier: str, id_short_path: str) -> bytes | None:
        """Downloads file content from a specific submodel element from the Submodel at a specified path. Experimental feature - may not be supported by all servers.

        :param submodel_identifier: The Submodels unique id
        :param id_short_path: IdShort path to the submodel element (dot-separated)
        :return: Attachment file data as bytes (octet-stream) or None if an error occurred
        """
        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodels/{submodel_identifier}/submodel-elements/{id_short_path}/attachment"

        self._client.set_token()  # ensures Authorization header is set

        try:
            response = self._session.get(url, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(
                    f"Submodel with id '{submodel_identifier}' or Submodel element with IDShort path '{id_short_path}' or file content not found."
                )
                _logger.debug(response.text)
                return None

            if response.status_code != STATUS_CODE_200:
                log_response(response)
                return None

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error calling REST API: {e}")
            return None

        return response.content

    # POST /submodels/{submodelIdentifier}/submodel-elements/{idShortPath}/attachment
    def post_file_by_path_submodel_repo(self, submodel_identifier: str, id_short_path: str, file: Path) -> bool:
        """Uploads file content to an existing submodel element at a specified path within submodel elements hierarchy. Experimental feature - may not be supported by all servers.

        :param submodel_identifier: The Submodels unique id
        :param id_short_path: IdShort path to the submodel element (dot-separated)
        :param file: Path to the file to upload as attachment
        :return: Attachment data as bytes or None if an error occurred
        """
        if file.exists() is False or not file.is_file():
            _logger.error(f"Attachment file '{file}' does not exist.")
            return False

        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodels/{submodel_identifier}/submodel-elements/{id_short_path}/attachment"

        self._client.set_token()

        try:
            mime_type, _ = mimetypes.guess_type(file)

            with file.open("rb") as f:
                files = {"file": (file.name, f, mime_type or "application/octet-stream")}
                response = self._session.post(url, files=files, timeout=self._client.time_out)

            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Submodel with id '{submodel_identifier}' or Submodel element with IDShort path '{id_short_path}' not found.")
                _logger.debug(response.text)
                return False

            # original dotnet server delivers 200 instead of 204
            if response.status_code not in (STATUS_CODE_200, STATUS_CODE_204):
                log_response(response)
                return False

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return False

        return True

    # PUT /submodels/{submodelIdentifier}/submodel-elements/{idShortPath}/attachment
    def put_file_by_path_submodel_repo(self, submodel_identifier: str, id_short_path: str, file: Path) -> bool:
        """Uploads file content to an existing submodel element at a specified path within submodel elements hierarchy. Experimental feature - may not be supported by all servers.

        :param submodel_identifier: The Submodels unique id
        :param id_short_path: IdShort path to the submodel element (dot-separated)
        :param file: Path to the file to upload as attachment
        :return: Attachment data as bytes or None if an error occurred
        """
        if file.exists() is False or not file.is_file():
            _logger.error(f"Attachment file '{file}' does not exist.")
            return False

        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodels/{submodel_identifier}/submodel-elements/{id_short_path}/attachment"

        self._client.set_token()

        try:
            mime_type, _ = mimetypes.guess_type(file)

            with file.open("rb") as f:
                files = {"file": (file.name, f, mime_type or "application/octet-stream")}
                response = self._session.put(url, files=files, timeout=self._client.time_out)

            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == STATUS_CODE_404:
                _logger.warning(f"Submodel with id '{submodel_identifier}' or Submodel element with IDShort path '{id_short_path}' not found.")
                _logger.debug(response.text)
                return False

            # original dotnet server delivers 200 instead of 204
            if response.status_code not in (STATUS_CODE_200, STATUS_CODE_204):
                log_response(response)
                return False

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error call REST API: {e}")
            return False

        return True

    # DELETE /submodels/{submodelIdentifier}/submodel-elements/{idShortPath}/attachment
    def delete_file_by_path_submodel_repo(self, submodel_identifier: str, id_short_path: str) -> bool:
        """Deletes file content of an existing submodel element at a specified path within submodel elements hierarchy. Experimental feature - may not be supported by all servers.

        :param submodel_identifier: The Submodels unique id
        :param id_short_path: IdShort path to the submodel element (dot-separated)
        :return: True if deletion was successful, False otherwise
        """
        if not self._client.encoded_ids:
            submodel_identifier = encode_base_64(submodel_identifier)

        url = f"{self._client.base_url}/submodels/{submodel_identifier}/submodel-elements/{id_short_path}/attachment"

        self._client.set_token()

        try:
            response = self._session.delete(url, timeout=self._client.time_out)
            _logger.debug(f"Call REST API url '{response.url}'")

            if response.status_code == 404:
                _logger.warning(f"Submodel with id '{submodel_identifier}' or Submodel element with IDShort path '{id_short_path}' not found.")
                return False

            if response.status_code != STATUS_CODE_200:
                log_response(response)
                return False

        except requests.exceptions.RequestException as e:
            _logger.error(f"Error calling REST API: {e}")
            return False

        return True

    def __post_multipart(self, url, files) -> requests.Response:
        headers = dict(self._session.headers)
        headers.pop("Content-Type", None)
        return self._session.post(url, headers=headers, files=files)



from basyx.aas import model
from aas_http_client.utilities import encoder
import aas_http_client.utilities.model_builder as model_builder
import aas_http_client.utilities.sdk_tools as sdk_tools
import pytest

SM_ID = "fluid40/sm_sdk_tools_unit_tests"
SHELL_ID = "fluid40/aas_sdk_tools_unit_tests"

@pytest.fixture(scope="module")
def shared_sm() -> model.Submodel:
    # create a Submodel
    return model_builder.create_base_submodel(identifier=SM_ID, id_short="sm_http_client_unit_tests", display_name="Submodel HTTP Client Unit Tests", description="This is a sample Submodel created for unit testing of the AAS HTTP Client.")

@pytest.fixture(scope="module")
def shared_sm_temp() -> model.Submodel:
    # create a Submodel
    return model_builder.create_base_submodel(identifier=f"{SM_ID}_temp", id_short="sm_http_client_unit_tests", display_name="Submodel HTTP Client Unit Tests", description="This is a sample Submodel created for unit testing of the AAS HTTP Client.")

@pytest.fixture(scope="module")
def shared_aas() -> model.AssetAdministrationShell:
    # create an AAS
    aas = model_builder.create_base_aas(identifier=SHELL_ID, id_short="aas_http_client_unit_tests", global_asset_identifier=SHELL_ID, display_name="AAS HTTP Client Unit Tests", description="This is a sample AAS created for unit testing of the AAS HTTP Client.")
    return aas

def test_001a_add_sm_to_shell(shared_aas: model.AssetAdministrationShell, shared_sm: model.Submodel):
    assert len(shared_aas.submodel) == 0

    sdk_tools.add_submodel_to_aas(shared_aas, shared_sm)
    assert len(shared_aas.submodel) == 1

    assert shared_sm.id in sdk_tools.get_submodel_ids(shared_aas)


def test_001b_add_sm_to_shell(shared_aas: model.AssetAdministrationShell, shared_sm: model.Submodel):
    assert len(shared_aas.submodel) == 1

    sdk_tools.add_submodel_to_aas(shared_aas, shared_sm)
    assert len(shared_aas.submodel) == 1

def test_002a_remove_sm_from_shell(shared_aas: model.AssetAdministrationShell, shared_sm_temp: model.Submodel):
    assert len(shared_aas.submodel) == 1

    sdk_tools.remove_submodel_from_aas(shared_aas, shared_sm_temp)
    assert len(shared_aas.submodel) == 1

def test_002b_remove_sm_from_shell(shared_aas: model.AssetAdministrationShell, shared_sm: model.Submodel):
    assert len(shared_aas.submodel) == 1

    sdk_tools.remove_submodel_from_aas(shared_aas, shared_sm)
    assert len(shared_aas.submodel) == 0
